# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline visualization.
"""

import io
import unittest

import lsst.pipe.base.connectionTypes as cT
import lsst.utils.tests
from lsst.pipe.base import Pipeline, PipelineTask, PipelineTaskConfig, PipelineTaskConnections
from lsst.pipe.base.mermaid_tools import pipeline2mermaid


class ExamplePipelineTaskConnections(PipelineTaskConnections, dimensions=("visit", "detector")):
    """Connections class used for testing.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        The config to use for this connections class.
    """

    input1 = cT.PrerequisiteInput(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExampleStorageClass",
        doc="Input for this task",
    )
    input2 = cT.Input(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExampleStorageClass",
        doc="Input for this task",
    )
    output1 = cT.Output(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExampleStorageClass",
        doc="Output for this task",
    )
    output2 = cT.Output(
        name="",
        dimensions=["visit", "detector"],
        storageClass="ExampleStorageClass",
        doc="Output for this task",
    )

    def __init__(self, *, config=None):
        super().__init__(config=config)
        if not config.connections.input2:
            self.inputs.remove("input2")
        if not config.connections.output2:
            self.outputs.remove("output2")


class ExamplePipelineTaskConfig(PipelineTaskConfig, pipelineConnections=ExamplePipelineTaskConnections):
    """Example config used for testing."""


def _makeConfig(inputName, outputName, pipeline, label):
    """Add config overrides.

    Factory method for config instances.

    inputName and outputName can be either string or tuple of strings
    with two items max.
    """
    if isinstance(inputName, tuple):
        pipeline.addConfigOverride(label, "connections.input1", inputName[0])
        pipeline.addConfigOverride(label, "connections.input2", inputName[1] if len(inputName) > 1 else "")
    else:
        pipeline.addConfigOverride(label, "connections.input1", inputName)

    if isinstance(outputName, tuple):
        pipeline.addConfigOverride(label, "connections.output1", outputName[0])
        pipeline.addConfigOverride(label, "connections.output2", outputName[1] if len(outputName) > 1 else "")
    else:
        pipeline.addConfigOverride(label, "connections.output1", outputName)


class ExamplePipelineTask(PipelineTask):
    """Example pipeline task used for testing."""

    ConfigClass = ExamplePipelineTaskConfig
    _DefaultName = "examplePipelineTask"


def _makePipeline(tasks):
    """Generate Pipeline instance.

    Parameters
    ----------
    tasks : list of tuples
        Each tuple in the list has 3 or 4 items:
        - input DatasetType name(s), string or tuple of strings
        - output DatasetType name(s), string or tuple of strings
        - task label, string
        - optional task class object, can be None

    Returns
    -------
    Pipeline instance
    """
    pipe = Pipeline("test pipeline")
    for task in tasks:
        inputs = task[0]
        outputs = task[1]
        label = task[2]
        klass = task[3] if len(task) > 3 else ExamplePipelineTask
        pipe.addTask(klass, label)
        _makeConfig(inputs, outputs, pipe, label)
    return list(pipe.to_graph()._iter_task_defs())


class MermaidToolsTestCase(unittest.TestCase):
    """A test case for Mermaid tools."""

    def test_pipeline2mermaid(self):
        """Tests for mermaid_tools.pipeline2mermaid method."""
        pipeline = _makePipeline(
            [
                ("A", ("B", "C"), "task0"),
                ("B", "E", "task1"),
                ("B", ("D", "F"), "task2"),
                ("D", "G", "task3"),
                (("C", "F"), "H", "task4"),
                ("task4_metadata", "I", "task5"),
            ]
        )
        file = io.StringIO()
        pipeline2mermaid(pipeline, file, expand_dimensions=True)

        # It's hard to validate complete output, just checking few basic
        # things, even that is not terribly stable.
        fileValue = file.getvalue()
        lines = fileValue.strip().split("\n")
        nClassDefs = 2
        nTasks = 6
        nTaskClass = 6
        nDatasets = 10
        nDatasetClass = 10
        nEdges = 16
        nLinkStyles = 2  # For the default and pre-requisite edges
        nExtra = 1  # 'flowchart' opening line

        self.assertEqual(
            len(lines),
            nClassDefs + nTasks + nTaskClass + nDatasets + nDatasetClass + nEdges + nLinkStyles + nExtra,
        )

        # Make sure the 'flowchart' definition is at the beginning of the file.
        self.assertTrue(fileValue.startswith("flowchart TD"))

        # Make sure components are connected appropriately.
        self.assertIn("DATASET_B --> TASK_1", fileValue)
        self.assertIn("TASK_2 --> DATASET_F", fileValue)

        # Make sure class definitions are created for datasets.
        self.assertIn("class DATASET_E ds;", fileValue)

        # Make sure there is a connection created for metadata if someone tries
        # to read it in.
        self.assertIn("TASK_4 --> DATASET_task4_metadata", fileValue)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Generic file handle leak check."""


def setup_module(module):
    """Set up the module for pytest.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module to set up.
    """
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining few methods to generate Mermaid charts from pipelines or
quantum graphs.
"""

from __future__ import annotations

__all__ = ["graph2mermaid", "pipeline2mermaid"]

import html
import re
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from lsst.daf.butler import DatasetType, DimensionUniverse

from . import connectionTypes
from .connections import iterConnections
from .pipeline import Pipeline

if TYPE_CHECKING:
    from lsst.daf.butler import DatasetRef
    from lsst.pipe.base import QuantumGraph, TaskDef


def _datasetRefId(dsRef: DatasetRef) -> str:
    """Make a unique identifier string for a dataset ref based on its name and
    dataId.
    """
    dsIdParts = [dsRef.datasetType.name]
    dsIdParts.extend(f"{key}_{dsRef.dataId[key]}" for key in sorted(dsRef.dataId.required.keys()))
    return "_".join(dsIdParts)


def _makeDatasetNode(dsRef: DatasetRef, allDatasetRefs: dict[str, str], file: Any) -> str:
    """Create a Mermaid node for a dataset if it doesn't exist, and return its
    node ID.
    """
    dsId = _datasetRefId(dsRef)
    nodeName = allDatasetRefs.get(dsId)
    if nodeName is None:
        nodeName = f"DATASET_{len(allDatasetRefs)}"
        allDatasetRefs[dsId] = nodeName
        # Simple label: datasetType name and run.
        label_lines = [f"**{dsRef.datasetType.name}**", f"run: {dsRef.run}"]
        # Add dataId info.
        for k in sorted(dsRef.dataId.required.keys()):
            label_lines.append(f"{k}={dsRef.dataId[k]}")
        label = "<br>".join(label_lines)
        print(f'{nodeName}["{label}"]', file=file)
    return nodeName


def graph2mermaid(qgraph: QuantumGraph, file: Any) -> None:
    """Convert QuantumGraph into a Mermaid flowchart (top-down).

    Parameters
    ----------
    qgraph : `~lsst.pipe.base.QuantumGraph`
        QuantumGraph instance.
    file : `str` or file object
        File where Mermaid flowchart is written, can be a file name or file
        object.

    Raises
    ------
    OSError
        Raised if the output file cannot be opened.
    ImportError
        Raised if the task class cannot be imported.
    """
    # Open a file if needed.
    close = False
    if not hasattr(file, "write"):
        file = open(file, "w")
        close = True

    # Start Mermaid code block with flowchart.
    print("flowchart TD", file=file)

    # To avoid duplicating dataset nodes, we track them.
    allDatasetRefs: dict[str, str] = {}

    # Process each task/quantum.
    for taskId, taskDef in enumerate(qgraph.taskGraph):
        quanta = qgraph.getNodesForTask(taskDef)
        for qId, quantumNode in enumerate(quanta):
            # Create quantum node.
            taskNodeName = f"TASK_{taskId}_{qId}"
            taskLabelLines = [f"**{taskDef.label}**", f"Node ID: {quantumNode.nodeId}"]
            dataId = quantumNode.quantum.dataId
            if dataId is not None:
                for k in sorted(dataId.required.keys()):
                    taskLabelLines.append(f"{k}={dataId[k]}")
            else:
                raise ValueError("Quantum DataId cannot be None")
            taskLabel = "<br>".join(taskLabelLines)
            print(f'{taskNodeName}["{taskLabel}"]', file=file)

            # Quantum inputs: datasets --> tasks
            for dsRefs in quantumNode.quantum.inputs.values():
                for dsRef in dsRefs:
                    dsNode = _makeDatasetNode(dsRef, allDatasetRefs, file)
                    print(f"{dsNode} --> {taskNodeName}", file=file)

            # Quantum outputs: tasks --> datasets
            for dsRefs in quantumNode.quantum.outputs.values():
                for dsRef in dsRefs:
                    dsNode = _makeDatasetNode(dsRef, allDatasetRefs, file)
                    print(f"{taskNodeName} --> {dsNode}", file=file)

    if close:
        file.close()


def _expand_dimensions(dimension_list: set[str] | Iterable[str], universe: DimensionUniverse) -> list[str]:
    """Return expanded list of dimensions, with special skypix treatment.

    Parameters
    ----------
    dimension_set : `set` [`str`] or iterable of `str`
        The original set of dimension names.
    universe : DimensionUniverse
        Used to conform the dimension set according to a known schema.

    Returns
    -------
    dimensions : `list` [`str`]
        Expanded list of dimensions.
    """
    dimension_set = set(dimension_list)
    skypix_dim = []
    if "skypix" in dimension_set:
        dimension_set.remove("skypix")
        skypix_dim = ["skypix"]
    dimensions = universe.conform(dimension_set)
    return list(dimensions.names) + skypix_dim


def _format_dimensions(dims: list[str]) -> str:
    """Format and sort dimension names as a comma-separated list inside curly
    braces.

    For example, if dims=["detector", "visit"], returns "{detector, visit}".

    Parameters
    ----------
    dims : list of str
        The dimension names to format and sort.

    Returns
    -------
    str
        The formatted dimension string, or an empty string if no dimensions.
    """
    if not dims:
        return ""
    sorted_dims = sorted(dims)
    return "{" + ",&nbsp;".join(sorted_dims) + "}"


def _render_task_node(
    task_id: str,
    taskDef: TaskDef,
    universe: DimensionUniverse,
    file: Any,
    show_dimensions: bool,
    expand_dimensions: bool,
) -> None:
    """Render a single task node in the Mermaid diagram.

    Parameters
    ----------
    task_id : str
        Unique Mermaid node identifier for this task.
    taskDef : TaskDef
        The pipeline task definition, which includes the task label, task name,
        and connections.
    universe : DimensionUniverse
        Used to conform and sort the task's dimensions.
    file : file-like
        The output file-like object to write the Mermaid node definition.
    show_dimensions : bool
        If True, display the task's dimensions after conforming them.
    expand_dimensions : bool
        If True, expand dimension names to include all components.
    """
    # Basic info: bold label, then task name.
    lines = [
        f"<b>{html.escape(taskDef.label)}</b>",
        html.escape(taskDef.taskName),
    ]

    # If requested, display the task's conformed dimensions.
    if show_dimensions and taskDef.connections and taskDef.connections.dimensions:
        if expand_dimensions:
            task_dims = _expand_dimensions(taskDef.connections.dimensions, universe)
        else:
            task_dims = list(taskDef.connections.dimensions)
        if task_dims:
            dim_str = _format_dimensions(task_dims)
            lines.append(f"<i>dimensions:</i>&nbsp;{dim_str}")

    # Join with <br> for line breaks and define the node with the label.
    label = "<br>".join(lines)
    print(f'{task_id}["{label}"]', file=file)
    print(f"class {task_id} task;", file=file)


def _render_dataset_node(
    ds_id: str,
    ds_name: str,
    connection: connectionTypes.BaseConnection,
    universe: DimensionUniverse,
    file: Any,
    show_dimensions: bool,
    expand_dimensions: bool,
    show_storage: bool,
) -> None:
    """Render a dataset-type node in the Mermaid diagram.

    Parameters
    ----------
    ds_id : str
        Unique Mermaid node identifier for this dataset.
    ds_name : str
        The dataset type name.
    connection : BaseConnection
        The dataset connection object, potentially dimensioned and having a
        storage class.
    universe : DimensionUniverse
        Used to conform and sort the dataset's dimensions if it is dimensioned.
    file : file-like
        The output file-like object to write the Mermaid node definition.
    show_dimensions : bool
        If True, display the dataset's conformed dimensions.
    expand_dimensions : bool
        If True, expand dimension names to include all components.
    show_storage : bool
        If True, display the dataset's storage class if available.
    """
    # Start with the dataset name in bold.
    lines = [f"<b>{html.escape(ds_name)}</b>"]

    # If dimensioned and requested, show conformed dimensions.
    ds_dims = []
    if show_dimensions and isinstance(connection, connectionTypes.DimensionedConnection):
        if expand_dimensions:
            ds_dims = _expand_dimensions(connection.dimensions, universe)
        else:
            ds_dims = list(connection.dimensions)

    if ds_dims:
        dim_str = _format_dimensions(ds_dims)
        lines.append(f"<i>dimensions:</i>&nbsp;{dim_str}")

    # If storage class is available and requested, display it.
    if show_storage and getattr(connection, "storageClass", None) is not None:
        lines.append(f"<i>storage&nbsp;class:</i>&nbsp;{html.escape(str(connection.storageClass))}")

    label = "<br>".join(lines)
    print(f'{ds_id}["{label}"]', file=file)
    print(f"class {ds_id} ds;", file=file)


def pipeline2mermaid(
    pipeline: Pipeline | Iterable[TaskDef],
    file: Any,
    show_dimensions: bool = True,
    expand_dimensions: bool = False,
    show_storage: bool = True,
) -> None:
    """Convert a Pipeline into a Mermaid flowchart diagram.

    This function produces a Mermaid flowchart, representing tasks and their
    inputs/outputs as dataset nodes. It uses a top-down layout.

    Parameters
    ----------
    pipeline : Pipeline or Iterable[TaskDef]
        The pipeline or collection of tasks to represent.
    file : str or file-like
        The output file or file-like object into which the Mermaid code is
        written.
    show_dimensions : bool, optional
        If True, display dimension information for tasks and datasets.
        Default is True.
    expand_dimensions : bool, optional
        If True, expand dimension names to include all components. Default is
        False.
    show_storage : bool, optional
        If True, display storage class information for datasets. Default is
        True.

    Raises
    ------
    OSError
        Raised if the output file cannot be opened.
    ImportError
        Raised if the task class cannot be imported.
    """
    universe = DimensionUniverse()

    # Ensure that pipeline is iterable of task definitions.
    if isinstance(pipeline, Pipeline):
        pipeline = pipeline.to_graph()._iter_task_defs()

    # Open file if needed.
    close = False
    if not hasattr(file, "write"):
        file = open(file, "w")
        close = True

    # Begin the Mermaid code block with top-down layout.
    print("flowchart TD", file=file)

    # Define classes for tasks and datasets.
    print(
        "classDef task fill:#B1F2EF,color:#000,stroke:#000,stroke-width:3px,"
        "font-family:Monospace,font-size:14px,text-align:left;",
        file=file,
    )
    print(
        "classDef ds fill:#F5F5F5,color:#000,stroke:#00BABC,stroke-width:3px,"
        "font-family:Monospace,font-size:14px,text-align:left,rx:10,ry:10;",
        file=file,
    )

    # Track which datasets have been rendered to avoid duplicates.
    allDatasets: set[str | tuple[str, str]] = set()

    # Used for linking metadata datasets after tasks are processed.
    labelToTaskName = {}
    metadataNodesToLink = set()

    # We'll store edges as (from_node, to_node, is_prerequisite) tuples.
    edges: list[tuple[str, str, bool]] = []

    def get_task_id(idx: int) -> str:
        """Generate a safe Mermaid node ID for a task."""
        return f"TASK_{idx}"

    def get_dataset_id(name: str) -> str:
        """Generate a safe Mermaid node ID for a dataset."""
        # Replace non-alphanumerics with underscores.
        return "DATASET_" + re.sub(r"[^0-9A-Za-z_]", "_", name)

    metadata_pattern = re.compile(r"^(.*)_metadata$")

    # Sort tasks by label for consistent diagram ordering.
    pipeline_tasks = sorted(pipeline, key=lambda x: x.label)

    # Process each task and its connections.
    for idx, taskDef in enumerate(pipeline_tasks):
        task_id = get_task_id(idx)
        labelToTaskName[taskDef.label] = task_id

        # Render the task node.
        _render_task_node(task_id, taskDef, universe, file, show_dimensions, expand_dimensions)

        # Handle standard inputs (non-prerequisite).
        for attr in sorted(iterConnections(taskDef.connections, "inputs"), key=lambda x: x.name):
            ds_id = get_dataset_id(attr.name)
            if attr.name not in allDatasets:
                _render_dataset_node(
                    ds_id, attr.name, attr, universe, file, show_dimensions, expand_dimensions, show_storage
                )
                allDatasets.add(attr.name)
            edges.append((ds_id, task_id, False))

            # Handle component datasets (composite -> component).
            nodeName, component = DatasetType.splitDatasetTypeName(attr.name)
            if component is not None and (nodeName, attr.name) not in allDatasets:
                ds_id_parent = get_dataset_id(nodeName)
                if nodeName not in allDatasets:
                    _render_dataset_node(
                        ds_id_parent,
                        nodeName,
                        attr,
                        universe,
                        file,
                        show_dimensions,
                        expand_dimensions,
                        show_storage,
                    )
                    allDatasets.add(nodeName)
                edges.append((ds_id_parent, ds_id, False))
                allDatasets.add((nodeName, attr.name))

            # If this is a metadata dataset, record it for linking later.
            if (match := metadata_pattern.match(attr.name)) is not None:
                matchTaskLabel = match.group(1)
                metadataNodesToLink.add((matchTaskLabel, attr.name))

        # Handle prerequisite inputs (to be drawn with a dashed line).
        for attr in sorted(iterConnections(taskDef.connections, "prerequisiteInputs"), key=lambda x: x.name):
            ds_id = get_dataset_id(attr.name)
            if attr.name not in allDatasets:
                _render_dataset_node(
                    ds_id, attr.name, attr, universe, file, show_dimensions, expand_dimensions, show_storage
                )
                allDatasets.add(attr.name)
            edges.append((ds_id, task_id, True))

            # If this is a metadata dataset, record it for linking later.
            if (match := metadata_pattern.match(attr.name)) is not None:
                matchTaskLabel = match.group(1)
                metadataNodesToLink.add((matchTaskLabel, attr.name))

        # Handle outputs (task -> dataset).
        for attr in sorted(iterConnections(taskDef.connections, "outputs"), key=lambda x: x.name):
            ds_id = get_dataset_id(attr.name)
            if attr.name not in allDatasets:
                _render_dataset_node(
                    ds_id, attr.name, attr, universe, file, show_dimensions, expand_dimensions, show_storage
                )
                allDatasets.add(attr.name)
            edges.append((task_id, ds_id, False))

    # Link metadata datasets after all tasks processed.
    for matchLabel, dsTypeName in metadataNodesToLink:
        if (result := labelToTaskName.get(matchLabel)) is not None:
            ds_id = get_dataset_id(dsTypeName)
            edges.append((result, ds_id, False))

    # Print all edges and track which are prerequisite.
    prereq_indices = []
    for i, (f, t, p) in enumerate(edges):
        print(f"{f} --> {t}", file=file)
        if p:
            prereq_indices.append(i)

    # Apply default edge style
    print("linkStyle default stroke:#000,stroke-width:1.5px,font-family:Monospace,font-size:14px;", file=file)

    # Apply dashed style for all prerequisite edges in one line.
    if prereq_indices:
        prereq_str = ",".join(str(i) for i in prereq_indices)
        print(f"linkStyle {prereq_str} stroke-dasharray:5;", file=file)

    if close:
        file.close()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline visualization.
"""

import io
import re
import unittest

import lsst.pipe.base.connectionTypes as cT
import lsst.utils.tests
from lsst.pipe.base import Pipeline, PipelineTask, PipelineTaskConfig, PipelineTaskConnections
from lsst.pipe.base.dot_tools import pipeline2dot


class ExamplePipelineTaskConnections(PipelineTaskConnections, dimensions=()):
    """Connections class used for testing.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        The config to use for this connections class.
    """

    input1 = cT.Input(
        name="", dimensions=["visit", "detector"], storageClass="example", doc="Input for this task"
    )
    input2 = cT.Input(
        name="", dimensions=["visit", "detector"], storageClass="example", doc="Input for this task"
    )
    output1 = cT.Output(
        name="", dimensions=["visit", "detector"], storageClass="example", doc="Output for this task"
    )
    output2 = cT.Output(
        name="", dimensions=["visit", "detector"], storageClass="example", doc="Output for this task"
    )

    def __init__(self, *, config=None):
        super().__init__(config=config)
        if not config.connections.input2:
            self.inputs.remove("input2")
        if not config.connections.output2:
            self.outputs.remove("output2")


class ExamplePipelineTaskConfig(PipelineTaskConfig, pipelineConnections=ExamplePipelineTaskConnections):
    """Example config used for testing."""


def _makeConfig(inputName, outputName, pipeline, label):
    """Add config overrides.

    Factory method for config instances.

    inputName and outputName can be either string or tuple of strings
    with two items max.
    """
    if isinstance(inputName, tuple):
        pipeline.addConfigOverride(label, "connections.input1", inputName[0])
        pipeline.addConfigOverride(label, "connections.input2", inputName[1] if len(inputName) > 1 else "")
    else:
        pipeline.addConfigOverride(label, "connections.input1", inputName)

    if isinstance(outputName, tuple):
        pipeline.addConfigOverride(label, "connections.output1", outputName[0])
        pipeline.addConfigOverride(label, "connections.output2", outputName[1] if len(outputName) > 1 else "")
    else:
        pipeline.addConfigOverride(label, "connections.output1", outputName)


class ExamplePipelineTask(PipelineTask):
    """Example pipeline task used for testing."""

    ConfigClass = ExamplePipelineTaskConfig


def _makePipeline(tasks):
    """Generate Pipeline instance.

    Parameters
    ----------
    tasks : list of tuples
        Each tuple in the list has 3 or 4 items:
        - input DatasetType name(s), string or tuple of strings
        - output DatasetType name(s), string or tuple of strings
        - task label, string
        - optional task class object, can be None

    Returns
    -------
    Pipeline instance
    """
    pipe = Pipeline("test pipeline")
    for task in tasks:
        inputs = task[0]
        outputs = task[1]
        label = task[2]
        klass = task[3] if len(task) > 3 else ExamplePipelineTask
        pipe.addTask(klass, label)
        _makeConfig(inputs, outputs, pipe, label)
    return list(pipe.to_graph()._iter_task_defs())


class DotToolsTestCase(unittest.TestCase):
    """A test case for dotTools."""

    def test_pipeline2dot(self):
        """Tests for dot_tools.pipeline2dot method."""
        pipeline = _makePipeline(
            [
                ("A", ("B", "C"), "task0"),
                ("C", "E", "task1"),
                ("B", "D", "task2"),
                (("D", "E"), "F", "task3"),
                ("D.C", "G", "task4"),
                ("task3_metadata", "H", "task5"),
            ]
        )
        file = io.StringIO()
        pipeline2dot(pipeline, file)

        # It's hard to validate complete output, just checking few basic
        # things, even that is not terribly stable.
        lines = file.getvalue().strip().split("\n")
        nglobals = 3
        ndatasets = 10
        ntasks = 6
        nedges = 16
        nextra = 2  # graph header and closing
        self.assertEqual(len(lines), nglobals + ndatasets + ntasks + nedges + nextra)

        # make sure that all node names are quoted
        nodeRe = re.compile(r"^([^ ]+) \[.+\];$")
        edgeRe = re.compile(r"^([^ ]+) *-> *([^ ]+);$")
        for line in lines:
            match = nodeRe.match(line)
            if match:
                node = match.group(1)
                if node not in ["graph", "node", "edge"]:
                    self.assertEqual(node[0] + node[-1], '""')
                    continue
            match = edgeRe.match(line)
            if match:
                for group in (1, 2):
                    node = match.group(group)
                    self.assertEqual(node[0] + node[-1], '""')
                continue

        # make sure components are connected appropriately
        self.assertIn('"D" -> "D.C"', file.getvalue())

        # make sure there is a connection created for metadata if someone
        # tries to read it in
        self.assertIn('"task3" -> "task3_metadata"', file.getvalue())


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Generic file handle leak check."""


def setup_module(module):
    """Set up the module for pytest.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module to set up.
    """
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["get_cli_subcommands"]

import click

from . import cmd


def get_cli_subcommands() -> list[click.Command]:
    """Return the location of the CLI command plugin definitions.

    Returns
    -------
    commands : `list` [ `click.Command` ]
        The command-line subcommands provided by this package.
    """
    return [getattr(cmd, c) for c in cmd.__all__]



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for PipelineTaskConnections.
"""

import unittest
import warnings

import lsst.pipe.base as pipeBase
import lsst.utils.tests
import pytest
from lsst.pex.config import Field


class TestConnectionsClass(unittest.TestCase):
    """Test connection classes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Test dimensions
        self.test_dims = ("a", "b")

    def testConnectionsDeclaration(self):
        """Tests the declaration of a Connections Class."""
        with pytest.raises(TypeError):
            # This should raise because this Connections class is created with
            # no dimensions
            class TestConnections(pipeBase.PipelineTaskConnections):
                pass

        with pytest.raises(TypeError):
            # This should raise because this Connections class is created with
            # out template defaults
            class TestConnectionsTemplate(pipeBase.PipelineTaskConnections, dimensions=self.test_dims):
                field = pipeBase.connectionTypes.Input(
                    doc="Test", name="{template}test", dimensions=self.test_dims, storageClass="Dummy"
                )

        # This declaration should raise no exceptions
        class TestConnectionsWithDimensions(pipeBase.PipelineTaskConnections, dimensions=self.test_dims):
            pass

        # This should not raise
        class TestConnectionsWithTemplate(
            pipeBase.PipelineTaskConnections,
            dimensions=self.test_dims,
            defaultTemplates={"template": "working"},
        ):
            field = pipeBase.connectionTypes.Input(
                doc="Test", name="{template}test", dimensions=self.test_dims, storageClass="Dummy"
            )

    def testConnectionsOnConnectionsClass(self):
        class TestConnections(pipeBase.PipelineTaskConnections, dimensions=self.test_dims):
            initInput1 = pipeBase.connectionTypes.InitInput(
                doc="Test Init input", name="init_input", storageClass="Dummy"
            )
            initInput2 = pipeBase.connectionTypes.InitInput(
                doc="Test Init input", name="init_input2", storageClass="Dummy"
            )

            initOutput1 = pipeBase.connectionTypes.InitOutput(
                doc="Test Init output", name="init_output1", storageClass="Dummy"
            )
            initOutput2 = pipeBase.connectionTypes.InitOutput(
                doc="Test Init output", name="init_output2", storageClass="Dummy"
            )

            input1 = pipeBase.connectionTypes.Input(
                doc="test input", name="input2", dimensions=self.test_dims, storageClass="Dummy"
            )
            input2 = pipeBase.connectionTypes.Input(
                doc="test input", name="input2", dimensions=self.test_dims, storageClass="Dummy"
            )

            prereqInputs1 = pipeBase.connectionTypes.PrerequisiteInput(
                doc="test input", name="pre_input1", dimensions=self.test_dims, storageClass="Dummy"
            )
            prereqInputs2 = pipeBase.connectionTypes.PrerequisiteInput(
                doc="test input", name="pre_input2", dimensions=self.test_dims, storageClass="Dummy"
            )

            output1 = pipeBase.connectionTypes.Output(
                doc="test output", name="output", dimensions=self.test_dims, storageClass="Dummy"
            )
            output2 = pipeBase.connectionTypes.Output(
                doc="test output", name="output", dimensions=self.test_dims, storageClass="Dummy"
            )

        self.assertEqual(TestConnections.initInputs, frozenset(["initInput1", "initInput2"]))
        self.assertEqual(TestConnections.initOutputs, frozenset(["initOutput1", "initOutput2"]))
        self.assertEqual(TestConnections.inputs, frozenset(["input1", "input2"]))
        self.assertEqual(TestConnections.prerequisiteInputs, frozenset(["prereqInputs1", "prereqInputs2"]))
        self.assertEqual(TestConnections.outputs, frozenset(["output1", "output2"]))

    def buildTestConnections(self):
        class TestConnectionsWithTemplate(
            pipeBase.PipelineTaskConnections,
            dimensions=self.test_dims,
            defaultTemplates={"template": "working"},
        ):
            field = pipeBase.connectionTypes.Input(
                doc="Test", name="{template}test", dimensions=self.test_dims, storageClass="Dummy"
            )
            field2 = pipeBase.connectionTypes.Output(
                doc="Test", name="field2Type", dimensions=self.test_dims, storageClass="Dummy", multiple=True
            )

            def adjustQuantum(self, datasetRefMap):
                if len(datasetRefMap.field) < 2:
                    raise ValueError("This connection should have more than one entry")

        class TestConfig(pipeBase.PipelineTaskConfig, pipelineConnections=TestConnectionsWithTemplate):
            pass

        config = TestConfig()
        config.connections.template = "fromConfig"
        config.connections.field2 = "field2FromConfig"

        connections = TestConnectionsWithTemplate(config=config)
        return connections

    def testConnectionsInstantiation(self):
        connections = self.buildTestConnections()
        self.assertEqual(connections.field.name, "fromConfigtest")
        self.assertEqual(connections.field2.name, "field2FromConfig")
        self.assertEqual(connections.allConnections["field"].name, "fromConfigtest")
        self.assertEqual(connections.allConnections["field2"].name, "field2FromConfig")

    def testBuildDatasetRefs(self):
        connections = self.buildTestConnections()

        mockQuantum = pipeBase.Struct(
            inputs={"fromConfigtest": ["a"]}, outputs={"field2FromConfig": ["b", "c"]}
        )

        inputRefs, outputRefs = connections.buildDatasetRefs(mockQuantum)
        self.assertEqual(inputRefs.field, "a")
        self.assertEqual(outputRefs.field2, ["b", "c"])

    def testAdjustQuantum(self):
        connections = self.buildTestConnections()
        mockQuantum = pipeBase.Struct(
            inputs={"fromConfigtest": ["a"]}, outputs={"field2FromConfig": ["b", "c"]}
        )
        inputRefs, outputRefs = connections.buildDatasetRefs(mockQuantum)
        with self.assertRaises(ValueError):
            connections.adjustQuantum(inputRefs)

    def testDimensionCheck(self):
        with self.assertRaises(TypeError):

            class TestConnectionsWithBrokenDimensionsStr(pipeBase.PipelineTask, dimensions={"a"}):
                pass

        with self.assertRaises(TypeError):

            class TestConnectionsWithBrokenDimensionsIter(pipeBase.PipelineTask, dimensions=2):
                pass

        with self.assertRaises(TypeError):
            pipeBase.connectionTypes.Output(
                Doc="mock doc", dimensions={"a"}, name="output", storageClass="mock"
            )

        with self.assertRaises(TypeError):
            pipeBase.connectionTypes.Output(Doc="mock doc", dimensions=1, name="output", storageClass="mock")

    def test_deprecation(self) -> None:
        """Test support for deprecating connections."""

        class TestConnections(
            pipeBase.PipelineTaskConnections,
            dimensions=self.test_dims,
            defaultTemplates={"t1": "dataset_type_1"},
            deprecatedTemplates={"t1": "Deprecated in v600, will be removed after v601."},
        ):
            input1 = pipeBase.connectionTypes.Input(
                doc="Docs for input1",
                name="input1_{t1}",
                storageClass="StructuredDataDict",
                deprecated="Deprecated in v50000, will be removed after v50001.",
            )

            def __init__(self, config):
                if config.drop_input1:
                    del self.input1

        class TestConfig(pipeBase.PipelineTaskConfig, pipelineConnections=TestConnections):
            drop_input1 = Field("Remove the 'input1' connection if True", dtype=bool, default=False)

        config = TestConfig()
        with self.assertWarns(FutureWarning):
            config.connections.input1 = "dataset_type_2"
        with self.assertWarns(FutureWarning):
            config.connections.t1 = "dataset_type_3"

        with self.assertWarns(FutureWarning):
            TestConnections(config=config)

        config.drop_input1 = True

        with warnings.catch_warnings():
            warnings.simplefilter("error", FutureWarning)
            TestConnections(config=config)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module which defines ConfigOverrides class and related methods.
"""
from __future__ import annotations

__all__ = ["ConfigOverrides"]

import ast
import contextlib
import inspect
from enum import Enum
from operator import attrgetter
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any

from lsst.resources import ResourcePath

from ._instrument import Instrument

if TYPE_CHECKING:
    from .pipelineIR import ParametersIR

OverrideTypes = Enum("OverrideTypes", "Value File Python Instrument")


class _FrozenSimpleNamespace(SimpleNamespace):
    """SimpleNamespace subclass which disallows setting after construction"""

    def __init__(self, **kwargs: Any) -> None:
        object.__setattr__(self, "_frozen", False)
        super().__init__(**kwargs)
        self._frozen = True

    def __setattr__(self, __name: str, __value: Any) -> None:
        if self._frozen:
            raise ValueError("Cannot set attributes on parameters")
        else:
            return super().__setattr__(__name, __value)


class ConfigExpressionParser(ast.NodeVisitor):
    """An expression parser that will be used to transform configuration
    strings supplied from the command line or a pipeline into a python
    object.

    This is roughly equivalent to ast.literal_parser, but with the ability to
    transform strings that are valid variable names into the value associated
    with the name. Variables that should be considered valid are supplied to
    the constructor as a dictionary that maps a string to its corresponding
    value.

    This class in an internal implementation detail, and should not be exposed
    outside this module.

    Parameters
    ----------
    namespace : `dict` of `str` to variable
        This is a mapping of strings corresponding to variable names, to the
        object that is associated with that name.
    """

    def __init__(self, namespace):
        self.variables = namespace

    def visit_Name(self, node):
        """Handle a node corresponding to a variable name.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        # If the id (name) of the variable is in the dictionary of valid names,
        # load and return the corresponding variable.
        if node.id in self.variables:
            return self.variables[node.id]
        # If the node does not correspond to a valid variable, turn the name
        # into a string, as the user likely intended it as such.
        return f"{node.id}"

    def visit_List(self, node):
        """Build a list out of the sub nodes when a list node is
        encountered.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        return [self.visit(elm) for elm in node.elts]

    def visit_Tuple(self, node):
        """Build a list out of the sub nodes and then turn it into a
        tuple.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        return tuple(self.visit_List(node))

    def visit_Constant(self, node):
        """Return constant from node.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        return node.value

    def visit_Dict(self, node):
        """Build dict out of component nodes if dict node encountered.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        return {self.visit(key): self.visit(value) for key, value in zip(node.keys, node.values, strict=True)}

    def visit_Set(self, node):
        """Build set out of node is set encountered.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        return {self.visit(el) for el in node.elts}

    def visit_UnaryOp(self, node):
        """Handle unary operators.

        This method is visited if the node is a unary operator. Currently
        The only operator we support is the negative (-) operator, all others
        are passed to generic_visit method.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        if isinstance(node.op, ast.USub):
            value = self.visit(node.operand)
            return -1 * value
        self.generic_visit(node)

    def generic_visit(self, node):
        """Handle other node types.

        This method is called for all other node types. It will just raise
        a value error, because this is a type of expression that we do not
        support.

        Parameters
        ----------
        node : `ast.Name`
            Node to use.
        """
        raise ValueError("Unable to parse string into literal expression")


class ConfigOverrides:
    """Defines a set of overrides to be applied to a task config.

    Overrides for task configuration need to be applied by activator when
    creating task instances. This class represents an ordered set of such
    overrides which activator receives from some source (e.g. command line
    or some other configuration).

    Methods
    -------
    addFileOverride(filename)
        Add overrides from a specified file.
    addValueOverride(field, value)
        Add override for a specific field.
    applyTo(config)
        Apply all overrides to a `config` instance.

    Notes
    -----
    Serialization support for this class may be needed, will add later if
    necessary.
    """

    def __init__(self) -> None:
        self._overrides: list[tuple[OverrideTypes, Any]] = []
        self._parameters: SimpleNamespace | None = None

    def addParameters(self, parameters: ParametersIR) -> None:
        """Add parameters which will be substituted when applying overrides.

        Parameters
        ----------
        parameters : `ParametersIR`
            Override parameters in the form as read from a Pipeline file.

        Notes
        -----
        This method may be called more than once, but each call will overwrite
        any previous parameter defined with the same name.
        """
        if self._parameters is None:
            self._parameters = SimpleNamespace()

        for key, value in parameters.mapping.items():
            setattr(self._parameters, key, value)

    def addFileOverride(self, filename):
        """Add overrides from a specified file.

        Parameters
        ----------
        filename : convertible to `~lsst.resources.ResourcePath`
            Path or URI to the override file.  All URI schemes supported by
            `~lsst.resources.ResourcePath` are supported.
        """
        self._overrides.append((OverrideTypes.File, ResourcePath(filename)))

    def addValueOverride(self, field, value):
        """Add override for a specific field.

        This method is not very type-safe as it is designed to support
        use cases where input is given as string, e.g. command line
        activators. If `value` has a string type and setting of the field
        fails with `TypeError` the we'll attempt `eval()` the value and
        set the field with that value instead.

        Parameters
        ----------
        field : str
            Fully-qualified field name.
        value : `~typing.Any`
            Value to be given to a filed.
        """
        self._overrides.append((OverrideTypes.Value, (field, value)))

    def addPythonOverride(self, python_snippet: str) -> None:
        """Add Overrides by running a snippit of python code against a config.

        Parameters
        ----------
        python_snippet : str
            A string which is valid python code to be executed. This is done
            with config as the only local accessible value.
        """
        self._overrides.append((OverrideTypes.Python, python_snippet))

    def addInstrumentOverride(self, instrument: Instrument, task_name: str) -> None:
        """Apply any overrides that an instrument has for a task.

        Parameters
        ----------
        instrument : `Instrument`
            An instrument instance which will apply configs.
        task_name : str
            The _DefaultName of a task associated with a config, used to look
            up overrides from the instrument.
        """
        self._overrides.append((OverrideTypes.Instrument, (instrument, task_name)))

    def _parser(self, value, configParser):
        # Exception probably means it is a specific user string such as a URI.
        # Let the value return as a string to attempt to continue to
        # process as a string, another error will be raised in downstream
        # code if that assumption is wrong
        with contextlib.suppress(Exception):
            value = configParser.visit(ast.parse(value, mode="eval").body)
        return value

    def applyTo(self, config):
        """Apply all overrides to a task configuration object.

        Parameters
        ----------
        config : `pex.Config`
            Configuration to apply to; modified in place.

        Raises
        ------
        `Exception` is raised if operations on configuration object fail.
        """
        # Look up a stack of variables people may be using when setting
        # configs. Create a dictionary that will be used akin to a namespace
        # for the duration of this function.
        localVars = {}
        # pull in the variables that are declared in module scope of the config
        mod = inspect.getmodule(config)
        localVars.update({k: v for k, v in mod.__dict__.items() if not k.startswith("__")})
        # put the supplied config in the variables dictionary
        localVars["config"] = config
        extraLocals = None

        # If any parameters are supplied add them to the variables dictionary
        if self._parameters is not None:
            # make a copy of the params and "freeze" it
            localParams = _FrozenSimpleNamespace(**vars(self._parameters))
            localVars["parameters"] = localParams
            extraLocals = {"parameters": localParams}

        # Create a parser for config expressions that may be strings
        configParser = ConfigExpressionParser(namespace=localVars)

        for otype, override in self._overrides:
            if otype is OverrideTypes.File:
                with override.open("r") as buffer:
                    config.loadFromStream(buffer, filename=override.ospath, extraLocals=extraLocals)
            elif otype is OverrideTypes.Value:
                field, value = override
                if isinstance(value, str):
                    value = self._parser(value, configParser)
                # checking for dicts and lists here is needed because {} and []
                # are valid yaml syntax so they get converted before hitting
                # this method, so we must parse the elements.
                #
                # The same override would remain a string if specified on the
                # command line, and is handled above.
                if isinstance(value, dict):
                    new = {}
                    for k, v in value.items():
                        if isinstance(v, str):
                            new[k] = self._parser(v, configParser)
                        else:
                            new[k] = v
                    value = new
                elif isinstance(value, list):
                    new = []
                    for v in value:
                        if isinstance(v, str):
                            new.append(self._parser(v, configParser))
                        else:
                            new.append(v)
                    value = new
                # The field might be a string corresponding to a attribute
                # hierarchy, attempt to split off the last field which
                # will then be set.
                parent, *child = field.rsplit(".", maxsplit=1)
                if child:
                    # This branch means there was a hierarchy, get the
                    # field to set, and look up the sub config for which
                    # it is to be set
                    finalField = child[0]
                    tmpConfig = attrgetter(parent)(config)
                else:
                    # There is no hierarchy, the config is the base config
                    # and the field is exactly what was passed in
                    finalField = parent
                    tmpConfig = config
                # set the specified config
                setattr(tmpConfig, finalField, value)

            elif otype is OverrideTypes.Python:
                # exec python string with the context of all vars known. This
                # both lets people use a var they know about (maybe a bit
                # dangerous, but not really more so than arbitrary python exec,
                # and there are so many other places to worry about security
                # before we start changing this) and any imports that are done
                # in a python block will be put into this scope. This means
                # other config setting branches can make use of these
                # variables.
                exec(override, None, localVars)
            elif otype is OverrideTypes.Instrument:
                instrument, name = override
                instrument.applyConfigOverrides(name, config)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["retrieve_artifacts_for_quanta"]

import logging

from lsst.daf.butler import QuantumBackedButler
from lsst.pipe.base import QuantumGraph
from lsst.resources import ResourcePath

_LOG = logging.getLogger(__name__)


def retrieve_artifacts_for_quanta(
    graph: str,
    repo: str,
    dest: str,
    transfer: str,
    preserve_path: bool,
    clobber: bool,
    qgraph_node_id: list[str],
    include_inputs: bool,
    include_outputs: bool,
) -> list[ResourcePath]:
    """Retrieve artifacts referenced in a graph and store locally.

    Parameters
    ----------
    graph : `str`
        URI string of the quantum graph.
    repo : `str`
        URI string of the Butler repo to use.
    dest : `str`
        URI string of the directory to write the artifacts.
    transfer : `str`
        Transfer mode to use when placing artifacts in the destination.
    preserve_path : `bool`
        If `True` the full datastore path will be retained within the
        destination directory, else only the filename will be used.
    clobber : `bool`
        If `True` allow transfers to overwrite files at the destination.
    qgraph_node_id : `tuple` [ `str` ]
        Quanta to extract.
    include_inputs : `bool`
        Whether to include input datasets in retrieval.
    include_outputs : `bool`
        Whether to include output datasets in retrieval.

    Returns
    -------
    paths : `list` [ `lsst.resources.ResourcePath` ]
        The paths to the artifacts that were written.
    """
    # Read graph into memory.
    nodes = qgraph_node_id or None
    qgraph = QuantumGraph.loadUri(graph, nodes=nodes)

    refs, datastore_records = qgraph.get_refs(
        include_inputs=include_inputs,
        include_init_inputs=include_inputs,
        include_outputs=include_outputs,
        include_init_outputs=include_outputs,
        conform_outputs=True,  # Need to look for predicted outputs with correct storage class.
    )

    # Get data repository definitions from the QuantumGraph; these can have
    # different storage classes than those in the quanta.
    dataset_types = {dstype.name: dstype for dstype in qgraph.registryDatasetTypes()}

    # Make QBB, its config is the same as output Butler.
    qbb = QuantumBackedButler.from_predicted(
        config=repo,
        predicted_inputs=[ref.id for ref in refs],
        predicted_outputs=[],
        dimensions=qgraph.universe,
        datastore_records=datastore_records,
        dataset_types=dataset_types,
    )

    paths = qbb.retrieve_artifacts(
        refs, dest, transfer=transfer, overwrite=clobber, preserve_path=preserve_path
    )
    return paths



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["InMemoryDatasetHandle"]

import dataclasses
from typing import Any, cast

from frozendict import frozendict
from lsst.daf.butler import (
    DataCoordinate,
    DataId,
    DimensionUniverse,
    StorageClass,
    StorageClassDelegate,
    StorageClassFactory,
)


# Use an empty dataID as a default.
def _default_dataId() -> DataCoordinate:
    return DataCoordinate.make_empty(DimensionUniverse())


@dataclasses.dataclass(frozen=True, init=False)
class InMemoryDatasetHandle:
    """An in-memory version of a `~lsst.daf.butler.DeferredDatasetHandle`.

    Parameters
    ----------
    inMemoryDataset : `~typing.Any`
        The dataset to be used by this handle.
    storageClass : `~lsst.daf.butler.StorageClass` or `None`, optional
        The storage class associated with the in-memory dataset. If `None`
        and if a storage class is needed, an attempt will be made to work one
        out from the underlying python type.
    parameters : `dict` [`str`, `~typing.Any`]
        Parameters to be used with `get`.
    dataId : `~lsst.daf.butler.DataId` or `None`, optional
        The dataId associated with this dataset. Only used for compatibility
        with the Butler implementation. Can be used for logging messages
        by calling code. If ``dataId`` is not specified, a default empty
        dataId will be constructed.
    copy : `bool`, optional
        Whether to copy on `get` or not.
    **kwargs : `~typing.Any`
        If ``kwargs`` are provided without specifying a ``dataId``, those
        parameters will be converted into a dataId-like entity.
    """

    _empty = DataCoordinate.make_empty(DimensionUniverse())

    def __init__(
        self,
        inMemoryDataset: Any,
        *,
        storageClass: StorageClass | str | None = None,
        parameters: dict[str, Any] | None = None,
        dataId: DataId | None = None,
        copy: bool = False,
        **kwargs: Any,
    ):
        object.__setattr__(self, "inMemoryDataset", inMemoryDataset)
        object.__setattr__(self, "storageClass", storageClass)
        object.__setattr__(self, "parameters", parameters)
        object.__setattr__(self, "copy", copy)
        # Need to be able to construct a dataId from kwargs for convenience.
        # This will not be a full DataCoordinate.
        if dataId is None:
            if kwargs:
                dataId = frozendict(kwargs)
            else:
                dataId = self._empty
        elif kwargs:
            if isinstance(dataId, DataCoordinate):
                dataId = DataCoordinate.standardize(kwargs, defaults=dataId, universe=dataId.universe)
            else:
                new = dict(dataId)
                new.update(kwargs)
                dataId = frozendict(new)
        object.__setattr__(self, "dataId", dataId)

    def get(
        self,
        *,
        component: str | None = None,
        parameters: dict | None = None,
        storageClass: str | StorageClass | None = None,
    ) -> Any:
        """Retrieve the dataset pointed to by this handle.

        This handle may be used multiple times, possibly with different
        parameters.

        Parameters
        ----------
        component : `str` or None
            If the deferred object is a component dataset type, this parameter
            may specify the name of the component to use in the get operation.
        parameters : `dict` or None
            The parameters argument will be passed to the butler get method.
            It defaults to `None`. If the value is not `None`, this `dict` will
            be merged with the parameters dict used to construct the
            `~lsst.daf.butler.DeferredDatasetHandle` class.
        storageClass : `~lsst.daf.butler.StorageClass` or `str`, optional
            The storage class to be used to override the Python type
            returned by this method. By default the returned type matches
            the type stored. Specifying a read `~lsst.daf.butler.StorageClass`
            can force a different type to be returned.
            This type must be compatible with the original type.

        Returns
        -------
        return : `object`
            The dataset pointed to by this handle. Whether this returns the
            original object or a copy is controlled by the ``copy`` property
            of the handle that is set at handle construction time.
            If the stored object is `None` this method always returns `None`
            regardless of any component request or parameters.

        Raises
        ------
        KeyError
            Raised if a component or parameters are used but no storage
            class can be found.
        """
        if self.inMemoryDataset is None:
            return None

        if self.parameters is not None:
            mergedParameters = self.parameters.copy()
            if parameters is not None:
                mergedParameters.update(parameters)
        elif parameters is not None:
            mergedParameters = parameters
        else:
            mergedParameters = {}

        returnStorageClass: StorageClass | None = None
        if storageClass:
            if isinstance(storageClass, str):
                factory = StorageClassFactory()
                returnStorageClass = factory.getStorageClass(storageClass)
            else:
                returnStorageClass = storageClass

        inMemoryDataset = self.inMemoryDataset

        if self.copy:
            # An optimization might be to defer copying until any components
            # and parameters have been applied. This can be a problem since
            # most component storage classes do not bother to define a
            # storage class delegate and the default delegate uses deepcopy()
            # which can fail if explicit support for deepcopy() is missing
            # or pickle does not work.
            # Copying will require a storage class be determined, which is
            # not normally required for the default case of no parameters and
            # no components.
            thisStorageClass = self._getStorageClass()
            try:
                delegate = thisStorageClass.delegate()
            except TypeError:
                # Try the default copy options if no delegate is available.
                delegate = StorageClassDelegate(thisStorageClass)

            inMemoryDataset = delegate.copy(inMemoryDataset)

        if component or mergedParameters:
            # This requires a storage class look up to locate the delegate
            # class.
            thisStorageClass = self._getStorageClass()

            # Parameters for derived components are applied against the
            # composite.
            if component in thisStorageClass.derivedComponents:
                # For some reason MyPy doesn't see the line above as narrowing
                # 'component' from 'str | None' to 'str'.
                component = cast(str, component)
                thisStorageClass.validateParameters(parameters)

                # Process the parameters (hoping this never modified the
                # original object).
                inMemoryDataset = thisStorageClass.delegate().handleParameters(
                    inMemoryDataset, mergedParameters
                )
                mergedParameters = {}  # They have now been used

                readStorageClass = thisStorageClass.derivedComponents[component]
            else:
                if component:
                    readStorageClass = thisStorageClass.components[component]
                else:
                    readStorageClass = thisStorageClass
                readStorageClass.validateParameters(mergedParameters)

            if component:
                inMemoryDataset = thisStorageClass.delegate().getComponent(inMemoryDataset, component)

            if mergedParameters:
                inMemoryDataset = readStorageClass.delegate().handleParameters(
                    inMemoryDataset, mergedParameters
                )

            if returnStorageClass:
                return returnStorageClass.coerce_type(inMemoryDataset)
            return inMemoryDataset
        else:
            # If there are no parameters or component requests the object
            # can be returned as is, but possibly with conversion.
            if returnStorageClass:
                return returnStorageClass.coerce_type(inMemoryDataset)
            return inMemoryDataset

    def _getStorageClass(self) -> StorageClass:
        """Return the relevant storage class.

        Returns
        -------
        storageClass : `~lsst.daf.butler.StorageClass`
            The storage class associated with this handle, or one derived
            from the python type of the stored object.

        Raises
        ------
        KeyError
            Raised if the storage class could not be found.
        """
        factory = StorageClassFactory()
        if self.storageClass:
            if isinstance(self.storageClass, str):
                return factory.getStorageClass(self.storageClass)
            else:
                return self.storageClass

        # Need to match python type.
        pytype = type(self.inMemoryDataset)
        return factory.findStorageClass(pytype)

    inMemoryDataset: Any
    """The object to store in this dataset handle for later retrieval.
    """

    dataId: DataCoordinate | frozendict
    """The `~lsst.daf.butler.DataCoordinate` associated with this dataset
    handle.
    """

    storageClass: StorageClass | str | None = None
    """The name of the `~lsst.daf.butler.StorageClass` associated with this
    dataset.

    If `None`, the storage class will be looked up from the factory.
    """

    parameters: dict | None = None
    """Optional parameters that may be used to specify a subset of the dataset
    to be loaded (`dict` or `None`).
    """

    copy: bool = False
    """Control whether a copy of the in-memory dataset is returned for every
    call to `get()`."""



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = (
    "QuantumGraphExecutionReport",
    "TaskExecutionReport",
    "DatasetTypeExecutionReport",
    "lookup_quantum_data_id",
)

import dataclasses
import itertools
import logging
import uuid
from collections.abc import Iterable, Mapping
from typing import Any

import networkx
import yaml
from lsst.daf.butler import Butler, DataCoordinate, DatasetRef, Quantum
from lsst.resources import ResourcePathExpression

from .graph import QuantumGraph


@dataclasses.dataclass
class DatasetTypeExecutionReport:
    """A report on the number of produced datasets as well as the status of
    missing datasets based on metadata.

    A `DatasetTypeExecutionReport` is created for each
    `~lsst.daf.butler.DatasetType` in a `TaskExecutionReport`.
    """

    failed: set[DatasetRef] = dataclasses.field(default_factory=set)
    """Datasets not produced because their quanta failed directly in this
    run (`set`).
    """

    not_produced: set[DatasetRef] = dataclasses.field(default_factory=set)
    """Missing datasets which were not produced by successful quanta.
    """

    blocked: set[DatasetRef] = dataclasses.field(default_factory=set)
    """Datasets not produced due to an upstream failure (`set`).
    """

    n_produced: int = 0
    """Count of datasets produced (`int`).
    """

    n_expected: int = 0
    """Count of datasets expected (`int`)
    """

    def to_summary_dict(self) -> dict[str, Any]:
        r"""Summarize the DatasetTypeExecutionReport in a dictionary.

        Returns
        -------
        summary_dict : `dict`
            A count of the datasets with each outcome; the number of
            produced, ``failed``, ``not_produced``, and ``blocked``
            `~lsst.daf.butler.DatasetType`\ s.
            See above for attribute descriptions.
        """
        return {
            "produced": self.n_produced,
            "failed": len(self.failed),
            "not_produced": len(self.not_produced),
            "blocked": len(self.blocked),
            "expected": self.n_expected,
        }


@dataclasses.dataclass
class TaskExecutionReport:
    """A report on the status and content of a task in an executed quantum
    graph.

    Use task metadata to identify and inspect failures and report on output
    datasets.

    See Also
    --------
    QuantumGraphExecutionReport : Quantum graph report.
    DatasetTypeExecutionReport : DatasetType report.
    """

    failed: dict[uuid.UUID, DatasetRef] = dataclasses.field(default_factory=dict)
    """A mapping from quantum data ID to log dataset reference for quanta that
    failed directly in this run (`dict`).
    """

    n_succeeded: int = 0
    """A count of successful quanta.

    This may include quanta that did not produce any datasets; ie, raised
    `NoWorkFound`.
    """

    n_expected: int = 0
    """A count of expected quanta.
    """

    blocked: dict[uuid.UUID, DataCoordinate] = dataclasses.field(default_factory=dict)
    """A mapping of data IDs of quanta that were not attempted due to an
    upstream failure (`dict`).
    """

    output_datasets: dict[str, DatasetTypeExecutionReport] = dataclasses.field(default_factory=dict)
    """Missing and produced outputs of each `~lsst.daf.butler.DatasetType`
    (`dict`).
    """

    def inspect_quantum(
        self,
        quantum_id: uuid.UUID,
        quantum: Quantum,
        status_graph: networkx.DiGraph,
        refs: Mapping[str, Mapping[uuid.UUID, DatasetRef]],
        metadata_name: str,
        log_name: str,
    ) -> None:
        """Inspect a quantum of a quantum graph and ascertain the status of
        each associated data product.

        Parameters
        ----------
        quantum_id : `uuid.UUID`
            Unique identifier for the quantum to inspect.
        quantum : `Quantum`
            The specific node of the quantum graph to be inspected.
        status_graph : `networkx.DiGraph`
            The quantum graph produced by
            `QuantumGraphExecutionReport.make_reports` which steps through the
            quantum graph of a run and logs the status of each quantum.
        refs : `~collections.abc.Mapping` [ `str`,\
                `~collections.abc.Mapping` [ `uuid.UUID`,\
                `~lsst.daf.butler.DatasetRef` ] ]
            The DatasetRefs of each of the DatasetTypes produced by the task.
            Includes initialization, intermediate and output data products.
        metadata_name : `str`
            The metadata dataset name for the node.
        log_name : `str`
            The name of the log files for the node.

        See Also
        --------
        QuantumGraphExecutionReport.make_reports : Make reports.
        """
        (metadata_ref,) = quantum.outputs[metadata_name]
        (log_ref,) = quantum.outputs[log_name]
        blocked = False
        if metadata_ref.id not in refs[metadata_name]:
            if any(
                status_graph.nodes[upstream_quantum_id]["failed"]
                for upstream_dataset_id in status_graph.predecessors(quantum_id)
                for upstream_quantum_id in status_graph.predecessors(upstream_dataset_id)
            ):
                assert quantum.dataId is not None
                self.blocked[quantum_id] = quantum.dataId
                blocked = True
            else:
                self.failed[quantum_id] = log_ref
                # note: log_ref may or may not actually exist
            failed = True
        else:
            failed = False
            self.n_succeeded += 1
        status_graph.nodes[quantum_id]["failed"] = failed

        # Now, loop over the datasets to make a DatasetTypeExecutionReport.
        for output_ref in itertools.chain.from_iterable(quantum.outputs.values()):
            if output_ref == metadata_ref or output_ref == log_ref:
                continue
            if (dataset_type_report := self.output_datasets.get(output_ref.datasetType.name)) is None:
                dataset_type_report = DatasetTypeExecutionReport()
                self.output_datasets[output_ref.datasetType.name] = dataset_type_report
            if output_ref.id not in refs[output_ref.datasetType.name]:
                if failed:
                    if blocked:
                        dataset_type_report.blocked.add(output_ref)
                    else:
                        dataset_type_report.failed.add(output_ref)
                else:
                    dataset_type_report.not_produced.add(output_ref)
            else:
                dataset_type_report.n_produced += 1
            dataset_type_report.n_expected += 1

    def to_summary_dict(
        self, butler: Butler, do_store_logs: bool = True, human_readable: bool = False
    ) -> dict[str, Any]:
        """Summarize the results of the TaskExecutionReport in a dictionary.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The Butler used for this report.
        do_store_logs : `bool`
            Store the logs in the summary dictionary.
        human_readable : `bool`
            Store more human-readable information to be printed out to the
            command-line.

        Returns
        -------
        summary_dict : `dict`
            A dictionary containing:

            - outputs: A dictionary summarizing the
              DatasetTypeExecutionReport for each DatasetType associated with
              the task
            - failed_quanta: A dictionary of quanta which failed and their
              dataIDs by quantum graph node id
            - n_quanta_blocked: The number of quanta which failed due to
              upstream failures.
            - n_succeded: The number of quanta which succeeded.

            And possibly, if human-readable is passed:

            - errors: A dictionary of data ids associated with each error
              message. If `human-readable` and `do_store_logs`, this is stored
              here. Otherwise, if `do_store_logs`, it is stored in
              `failed_quanta` keyed by the quantum graph node id.
        """
        failed_quanta = {}
        failed_data_ids = []
        errors = []
        for node_id, log_ref in self.failed.items():
            data_id = dict(log_ref.dataId.required)
            quantum_info: dict[str, Any] = {"data_id": data_id}
            if do_store_logs:
                try:
                    log = butler.get(log_ref)
                except LookupError:
                    quantum_info["error"] = []
                except FileNotFoundError:
                    quantum_info["error"] = None
                else:
                    quantum_info["error"] = [
                        record.message for record in log if record.levelno >= logging.ERROR
                    ]
            if human_readable:
                failed_data_ids.append(data_id)
                if do_store_logs:
                    errors.append(quantum_info)

            else:
                failed_quanta[str(node_id)] = quantum_info
        result = {
            "outputs": {name: r.to_summary_dict() for name, r in self.output_datasets.items()},
            "n_quanta_blocked": len(self.blocked),
            "n_succeeded": self.n_succeeded,
            "n_expected": self.n_expected,
        }
        if human_readable:
            result["failed_quanta"] = failed_data_ids
            result["errors"] = errors
        else:
            result["failed_quanta"] = failed_quanta
        return result

    def __str__(self) -> str:
        """Return a count of the failed and blocked tasks in the
        TaskExecutionReport.
        """
        return f"failed: {len(self.failed)}\nblocked: {len(self.blocked)}\n"


@dataclasses.dataclass
class QuantumGraphExecutionReport:
    """A report on the execution of a quantum graph.

    Report the detailed status of each failure; whether tasks were not run,
    data is missing from upstream failures, or specific errors occurred during
    task execution (and report the errors). Contains a count of expected,
    produced DatasetTypes for each task. This report can be output as a
    dictionary or a yaml file.

    Attributes
    ----------
    tasks : `dict`
        A dictionary of TaskExecutionReports by task label.

    See Also
    --------
    TaskExecutionReport : A task report.
    DatasetTypeExecutionReport : A dataset type report.
    """

    tasks: dict[str, TaskExecutionReport] = dataclasses.field(default_factory=dict)
    """A dictionary of TaskExecutionReports by task label (`dict`)."""

    def to_summary_dict(
        self, butler: Butler, do_store_logs: bool = True, human_readable: bool = False
    ) -> dict[str, Any]:
        """Summarize the results of the `QuantumGraphExecutionReport` in a
        dictionary.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The Butler used for this report.
        do_store_logs : `bool`
            Store the logs in the summary dictionary.
        human_readable : `bool`
            Store more human-readable information to be printed out to the
            command-line.

        Returns
        -------
        summary_dict : `dict`
            A dictionary containing a summary of a `TaskExecutionReport` for
            each task in the quantum graph.
        """
        return {
            task: report.to_summary_dict(butler, do_store_logs=do_store_logs, human_readable=human_readable)
            for task, report in self.tasks.items()
        }

    def write_summary_yaml(self, butler: Butler, filename: str, do_store_logs: bool = True) -> None:
        """Take the dictionary from
        `QuantumGraphExecutionReport.to_summary_dict` and store its contents in
        a yaml file.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The Butler used for this report.
        filename : `str`
            The name to be used for the summary yaml file.
        do_store_logs : `bool`
            Store the logs in the summary dictionary.
        """
        with open(filename, "w") as stream:
            yaml.safe_dump(self.to_summary_dict(butler, do_store_logs=do_store_logs), stream)

    @classmethod
    def make_reports(
        cls,
        butler: Butler,
        graph: QuantumGraph | ResourcePathExpression,
    ) -> QuantumGraphExecutionReport:
        """Make a `QuantumGraphExecutionReport`.

        Step through the quantum graph associated with a run, creating a
        `networkx.DiGraph` called status_graph to annotate the status of each
        quantum node. For each task in the quantum graph, use
        `TaskExecutionReport.inspect_quantum` to make a `TaskExecutionReport`
        based on the status of each node. Return a `TaskExecutionReport` for
        each task in the quantum graph.

        Parameters
        ----------
        butler : `lsst.daf.butler.Butler`
            The Butler used for this report. This should match the Butler used
            for the run associated with the executed quantum graph.
        graph : `QuantumGraph` | `ResourcePathExpression`
            Either the associated quantum graph object or the uri of the
            location of said quantum graph.

        Returns
        -------
        report: `QuantumGraphExecutionReport`
            The `TaskExecutionReport` for each task in the quantum graph.
        """
        refs = {}  # type: dict[str, Any]
        status_graph = networkx.DiGraph()
        if not isinstance(graph, QuantumGraph):
            qg = QuantumGraph.loadUri(graph)
        else:
            qg = graph
        assert qg.metadata is not None, "Saved QGs always have metadata."
        collection = qg.metadata["output_run"]
        report = cls()
        for dataset_type_node in qg.pipeline_graph.dataset_types.values():
            if qg.pipeline_graph.producer_of(dataset_type_node.name) is None:
                continue
            refs[dataset_type_node.name] = {
                ref.id: ref
                for ref in butler.registry.queryDatasets(
                    dataset_type_node.name, collections=collection, findFirst=False
                )
            }
        for task_node in qg.pipeline_graph.tasks.values():
            for quantum_id, quantum in qg.get_task_quanta(task_node.label).items():
                status_graph.add_node(quantum_id)
                for ref in itertools.chain.from_iterable(quantum.outputs.values()):
                    status_graph.add_edge(quantum_id, ref.id)
                for ref in itertools.chain.from_iterable(quantum.inputs.values()):
                    status_graph.add_edge(ref.id, quantum_id)
        for task_node in qg.pipeline_graph.tasks.values():
            task_report = TaskExecutionReport()
            if task_node.log_output is None:
                raise RuntimeError("QG must have log outputs to use execution reports.")
            for quantum_id, quantum in qg.get_task_quanta(task_node.label).items():
                task_report.inspect_quantum(
                    quantum_id,
                    quantum,
                    status_graph,
                    refs,
                    metadata_name=task_node.metadata_output.dataset_type_name,
                    log_name=task_node.log_output.dataset_type_name,
                )
                task_report.n_expected = len(qg.get_task_quanta(task_node.label).items())
            report.tasks[task_node.label] = task_report
        return report

    def __str__(self) -> str:
        return "\n".join(f"{tasklabel}:{report}" for tasklabel, report in self.tasks.items())


def lookup_quantum_data_id(
    graph_uri: ResourcePathExpression, nodes: Iterable[uuid.UUID]
) -> list[DataCoordinate | None]:
    """Look up a dataId from a quantum graph and a list of quantum graph
    nodeIDs.

    Parameters
    ----------
    graph_uri : `ResourcePathExpression`
        URI of the quantum graph of the run.
    nodes : `~collections.abc.Iterable` [ `uuid.UUID` ]
        Quantum graph nodeID.

    Returns
    -------
    data_ids : `list` [ `lsst.daf.butler.DataCoordinate` ]
        A list of human-readable dataIDs which map to the nodeIDs on the
        quantum graph at graph_uri.
    """
    qg = QuantumGraph.loadUri(graph_uri, nodes=nodes)
    return [qg.getQuantumNodeByNodeId(node).quantum.dataId for node in nodes]



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("show_mermaid",)

import html
import os
import sys
from collections.abc import Mapping
from typing import Any, TextIO

from .._nodes import NodeType
from .._pipeline_graph import PipelineGraph
from ._formatting import NodeKey, format_dimensions, format_task_class
from ._options import NodeAttributeOptions
from ._show import parse_display_args

# Configuration constants for label formatting and overflow handling.
_LABEL_PX_SIZE = 18
_LABEL_MAX_LINES_SOFT = 10
_LABEL_MAX_LINES_HARD = 15
_OVERFLOW_MAX_LINES = 20


def show_mermaid(
    pipeline_graph: PipelineGraph,
    stream: TextIO = sys.stdout,
    **kwargs: Any,
) -> None:
    """Write a Mermaid flowchart representation of the pipeline graph to a
    stream.

    This function converts a given `PipelineGraph` into a Mermaid-based
    flowchart. Nodes represent tasks (and possibly task-init nodes) and dataset
    types, and edges represent connections between them. Dimensions and storage
    classes can be included as additional metadata on nodes. Prerequisite edges
    are rendered as dashed lines.

    Parameters
    ----------
    pipeline_graph : `PipelineGraph`
        The pipeline graph to visualize.
    stream : `TextIO`, optional
        The output stream where Mermaid code is written. Defaults to
        `sys.stdout`.
    **kwargs : Any
        Additional arguments passed to `parse_display_args` to control aspects
        such as displaying dimensions, storage classes, or full task class
        names.

    Notes
    -----
    - The diagram uses a top-down layout (`flowchart TD`).
    - Three Mermaid classes are defined:
      - `task` for normal tasks,
      - `dsType` for dataset-type nodes,
      - `taskInit` for task-init nodes.
    - Edges that represent prerequisite relationships are rendered as dashed
      lines using `linkStyle`.
    - If a node's label is too long, overflow nodes are created to hold extra
      lines.
    """
    # Parse display arguments to determine what to show.
    xgraph, options = parse_display_args(pipeline_graph, **kwargs)

    # Begin the Mermaid code block.
    print("flowchart TD", file=stream)

    # Define Mermaid classes for node styling.
    print(
        f"classDef task fill:#B1F2EF,color:#000,stroke:#000,stroke-width:3px,"
        f"font-family:Monospace,font-size:{_LABEL_PX_SIZE}px,text-align:left;",
        file=stream,
    )
    print(
        f"classDef dsType fill:#F5F5F5,color:#000,stroke:#00BABC,stroke-width:3px,"
        f"font-family:Monospace,font-size:{_LABEL_PX_SIZE}px,text-align:left,rx:8,ry:8;",
        file=stream,
    )
    print(
        f"classDef taskInit fill:#F4DEFA,color:#000,stroke:#000,stroke-width:3px,"
        f"font-family:Monospace,font-size:{_LABEL_PX_SIZE}px,text-align:left;",
        file=stream,
    )

    # `overflow_ref` tracks the reference numbers for overflow nodes.
    overflow_ref = 1
    overflow_ids = []

    # Render nodes.
    for node_key, node_data in xgraph.nodes.items():
        match node_key.node_type:
            case NodeType.TASK | NodeType.TASK_INIT:
                # Render a task or task-init node.
                _render_task_node(node_key, node_data, options, stream)
            case NodeType.DATASET_TYPE:
                # Render a dataset-type node with possible overflow handling.
                overflow_ref, node_overflow_ids = _render_dataset_type_node(
                    node_key, node_data, options, stream, overflow_ref
                )
                if node_overflow_ids:
                    overflow_ids += node_overflow_ids
            case _:
                raise AssertionError(f"Unexpected node type: {node_key.node_type}")

    # Collect edges for printing and track which ones are prerequisite
    # so we can apply dashed styling after printing them.
    edges = []
    for _, (from_node, to_node, *_rest) in enumerate(xgraph.edges):
        is_prereq = xgraph.nodes[from_node].get("is_prerequisite", False)
        edges.append((from_node.node_id, to_node.node_id, is_prereq))

    # Print all edges
    for _, (f, t, p) in enumerate(edges):
        _render_edge(f, t, p, stream)

    # After printing all edges, apply linkStyle to prerequisite edges to make
    # them dashed:

    # First, gather indices of prerequisite edges.
    prereq_indices = [str(i) for i, (_, _, p) in enumerate(edges) if p]

    # Then apply dashed styling to all prerequisite edges in one line.
    if prereq_indices:
        print(f"linkStyle {','.join(prereq_indices)} stroke-dasharray:5;", file=stream)


def _render_task_node(
    node_key: NodeKey,
    node_data: Mapping[str, Any],
    options: NodeAttributeOptions,
    stream: TextIO,
) -> None:
    """Render a Mermaid node for a task or task-init node.

    Parameters
    ----------
    node_key : NodeKey
        Identifies the node. The node type determines styling and whether
        dimensions apply.
    node_data : Mapping[str, Any]
        Node attributes, including possibly 'task_class_name' and 'dimensions'.
    options : NodeAttributeOptions
        Rendering options controlling whether to show dimensions, storage
        classes, etc.
    stream : TextIO
        The output stream for Mermaid syntax.
    """
    # Convert node_key into a label, handling line splitting and prefix
    # extraction.
    lines, _, _ = _format_label(str(node_key))

    # If requested, show the fully qualified task class name beneath the task
    # label.
    if options.task_classes and node_key.node_type in (NodeType.TASK, NodeType.TASK_INIT):
        lines.append(html.escape(format_task_class(options, node_data["task_class_name"])))

    # Show dimensions if requested and if this is not a task-init node.
    if options.dimensions and node_key.node_type != NodeType.TASK_INIT:
        dims_str = html.escape(format_dimensions(options, node_data["dimensions"])).replace(" ", "&nbsp;")
        lines.append(f"<i>dimensions:</i>&nbsp;{dims_str}")

    # Join lines with <br> for multi-line label.
    label = "<br>".join(lines)

    # Print Mermaid node.
    node_id = node_key.node_id
    print(f'{node_id}["{label}"]', file=stream)

    # Assign class based on node type.
    if node_key.node_type == NodeType.TASK:
        print(f"class {node_id} task;", file=stream)
    else:
        # For NodeType.TASK_INIT.
        print(f"class {node_id} taskInit;", file=stream)


def _render_dataset_type_node(
    node_key: NodeKey,
    node_data: Mapping[str, Any],
    options: NodeAttributeOptions,
    stream: TextIO,
    overflow_ref: int,
) -> tuple[int, list[str]]:
    """Render a Mermaid node for a dataset-type node, handling overflow lines
    if needed.

    Dataset-type nodes can have many lines of label text. If the label exceeds
    a certain threshold, we create separate "overflow" nodes.

    Parameters
    ----------
    node_key : NodeKey
        Identifies this dataset-type node.
    node_data : Mapping[str, Any]
        Node attributes, possibly including dimensions and storage class.
    options : NodeAttributeOptions
        Rendering options controlling whether to show dimensions and storage
        classes.
    stream : TextIO
        The output stream for Mermaid syntax.
    overflow_ref : int
        The current reference number for overflow nodes. If overflow occurs,
        this is incremented.

    Returns
    -------
    overflow_ref : int
        Possibly incremented overflow reference number.
    overflow_ids : list[str]
        IDs of overflow nodes created, if any.
    """
    # Format the node label, respecting soft/hard line limits.
    labels, label_extras, _ = _format_label(str(node_key), _LABEL_MAX_LINES_SOFT)

    overflow_ids = []
    total_lines = len(labels) + len(label_extras)
    if total_lines > _LABEL_MAX_LINES_HARD:
        # Too many lines, we must handle overflow by splitting extras.
        allowed_extras = _LABEL_MAX_LINES_HARD - len(labels)
        if allowed_extras < 0:
            allowed_extras = 0
        extras_for_overflow = label_extras[allowed_extras:]
        label_extras = label_extras[:allowed_extras]

        if extras_for_overflow:
            # Introduce an overflow anchor.
            overflow_anchor = f"[{overflow_ref}]"
            labels.append(f"<b>...more details in {overflow_anchor}</b>")

            # Create overflow nodes in chunks.
            for i in range(0, len(extras_for_overflow), _OVERFLOW_MAX_LINES):
                overflow_id = f"{node_key.node_id}_overflow_{overflow_ref}_{i}"
                chunk = extras_for_overflow[i : i + _OVERFLOW_MAX_LINES]
                chunk.insert(0, f"<b>{html.escape(overflow_anchor)}</b>")
                _render_simple_node(overflow_id, chunk, "dsType", stream)
                overflow_ids.append(overflow_id)

            overflow_ref += 1

    # Combine final lines after overflow handling.
    final_lines = labels + label_extras

    # Append dimensions if requested.
    if options.dimensions:
        dims_str = html.escape(format_dimensions(options, node_data["dimensions"])).replace(" ", "&nbsp;")
        final_lines.append(f"<i>dimensions:</i>&nbsp;{dims_str}")

    # Append storage class if requested.
    if options.storage_classes:
        final_lines.append(f"<i>storage&nbsp;class:</i>&nbsp;{html.escape(node_data['storage_class_name'])}")

    # Render the main dataset-type node.
    _render_simple_node(node_key.node_id, final_lines, "dsType", stream)

    return overflow_ref, overflow_ids


def _render_simple_node(node_id: str, lines: list[str], node_class: str, stream: TextIO) -> None:
    """Render a simple Mermaid node with given lines and a class.

    This helper function is used for both primary nodes and overflow nodes once
    the split has been decided.

    Parameters
    ----------
    node_id : str
        Mermaid node ID.
    lines : list[str]
        Lines of HTML-formatted text to display in the node.
    node_class : str
        Mermaid class name to style the node (e.g., 'dsType', 'task',
        'taskInit').
    stream : TextIO
        The output stream.
    """
    label = "<br>".join(lines)
    print(f'{node_id}["{label}"]', file=stream)
    print(f"class {node_id} {node_class};", file=stream)


def _render_edge(from_node_id: str, to_node_id: str, is_prerequisite: bool, stream: TextIO) -> None:
    """Render a Mermaid edge from one node to another.

    Edges in Mermaid are normally specified as `A --> B`. Prerequisite edges
    will later be styled as dashed lines using linkStyle after all edges have
    been printed.

    Parameters
    ----------
    from_node_id : str
        The ID of the 'from' node in the edge.
    to_node_id : str
        The ID of the 'to' node in the edge.
    is_prerequisite : bool
        If True, this edge represents a prerequisite connection and will be
        styled as dashed.
    stream : TextIO
        The output stream for Mermaid syntax.
    """
    # At this stage, we simply print the edge. The styling (dashed) for
    # prerequisite edges is applied afterwards via linkStyle lines.
    print(f"{from_node_id} --> {to_node_id}", file=stream)


def _format_label(
    label: str,
    max_lines: int = 10,
    min_common_prefix_len: int = 1000,
) -> tuple[list[str], list[str], str]:
    """Parse and format a label into multiple lines with optional overflow
    handling.

    This function attempts to cleanly format long labels by:
    - Splitting the label by ", ".
    - Identifying a common prefix to factor out if sufficiently long.
    - Limiting the number of lines to 'max_lines', storing extras for potential
      overflow.

    Parameters
    ----------
    label : str
        The raw label text, often derived from a NodeKey.
    max_lines : int, optional
        Maximum lines before overflow is triggered.
    min_common_prefix_len : int, optional
        Minimum length for considering a common prefix extraction.

    Returns
    -------
    labels : list[str]
        Main label lines as HTML-formatted text.
    label_extras : list[str]
        Overflow lines if the label is too long.
    common_prefix : str
        Extracted common prefix, if any.
    """
    parsed_labels, parsed_label_extras, common_prefix = _parse_label(label, max_lines, min_common_prefix_len)

    # If there's a common prefix, present it bolded.
    if common_prefix:
        common_prefix = f"<b>{html.escape(common_prefix)}:</b>"

    indent = "&nbsp;&nbsp;" if common_prefix else ""
    labels = [f"<b>{indent}{html.escape(el)}</b>" for el in parsed_labels]
    label_extras = [f"<b>{indent}{html.escape(el)}</b>" for el in parsed_label_extras]

    if common_prefix:
        labels.insert(0, common_prefix)

    return labels, label_extras, common_prefix or ""


def _parse_label(
    label: str,
    max_lines: int,
    min_common_prefix_len: int,
) -> tuple[list[str], list[str], str]:
    """Split and process label text for overflow and common prefix extraction.

    Parameters
    ----------
    label : str
        The raw label text.
    max_lines : int
        Maximum number of lines before overflow.
    min_common_prefix_len : int
        Minimum length for a common prefix to be considered.

    Returns
    -------
    labels : list[str]
        The primary label lines.
    label_extras : list[str]
        Any overflow lines that exceed max_lines.
    common_prefix : str
        The extracted common prefix, if applicable.
    """
    labels = label.split(", ")
    common_prefix = os.path.commonprefix(labels)

    # If there's a long common prefix for multiple labels, factor it out at the
    # nearest underscore.
    if len(labels) > 3 and len(common_prefix) > min_common_prefix_len:
        final_underscore_index = common_prefix.rfind("_")
        if final_underscore_index > 0:
            common_prefix = common_prefix[: final_underscore_index + 1]
            labels = [element[len(common_prefix) :] for element in labels]
        else:
            common_prefix = ""
    else:
        common_prefix = ""

    # Handle overflow if needed.
    if (len(labels) + bool(common_prefix)) > max_lines:
        label_extras = labels[max_lines - bool(common_prefix) :]
        labels = labels[: max_lines - bool(common_prefix)]
    else:
        label_extras = []

    return labels, label_extras, common_prefix



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Tests of things related to the GraphBuilder class."""

import io
import logging
import unittest

import lsst.utils.tests
from lsst.daf.butler.registry import UserExpressionError
from lsst.pipe.base import QuantumGraph
from lsst.pipe.base.all_dimensions_quantum_graph_builder import DatasetQueryConstraintVariant
from lsst.pipe.base.tests import simpleQGraph
from lsst.utils.tests import temporaryDirectory

_LOG = logging.getLogger(__name__)


class GraphBuilderTestCase(unittest.TestCase):
    """Test graph building."""

    def _assertGraph(self, graph: QuantumGraph) -> None:
        """Check basic structure of the graph."""
        for taskDef in graph.iterTaskGraph():
            refs = graph.initOutputRefs(taskDef)
            # task has one initOutput, second ref is for config dataset
            self.assertEqual(len(refs), 2)

        self.assertEqual(len(list(graph.inputQuanta)), 1)

        # This includes only "packages" dataset for now.
        refs = graph.globalInitOutputRefs()
        self.assertEqual(len(refs), 1)

    def testDefault(self):
        """Simple test to verify makeSimpleQGraph can be used to make a Quantum
        Graph.
        """
        with temporaryDirectory() as root:
            # makeSimpleQGraph calls GraphBuilder.
            butler, qgraph = simpleQGraph.makeSimpleQGraph(root=root)
            # by default makeSimpleQGraph makes a graph with 5 nodes
            self.assertEqual(len(qgraph), 5)
            self._assertGraph(qgraph)
            constraint = DatasetQueryConstraintVariant.OFF
            _, qgraph2 = simpleQGraph.makeSimpleQGraph(
                butler=butler, datasetQueryConstraint=constraint, callPopulateButler=False
            )
            # When all outputs are random resolved refs, direct comparison
            # of graphs does not work because IDs are different. Can only
            # verify the number of quanta in the graph without doing something
            # terribly complicated.
            self.assertEqual(len(qgraph2), 5)
            constraint = DatasetQueryConstraintVariant.fromExpression("add_dataset0")
            _, qgraph3 = simpleQGraph.makeSimpleQGraph(
                butler=butler, datasetQueryConstraint=constraint, callPopulateButler=False
            )
            self.assertEqual(len(qgraph3), 5)

    def testAddInstrumentMismatch(self):
        """Verify that a RuntimeError is raised if the instrument in the user
        query does not match the instrument in the pipeline.
        """
        with temporaryDirectory() as root:
            pipeline = simpleQGraph.makeSimplePipeline(
                nQuanta=5, instrument="lsst.pipe.base.tests.simpleQGraph.SimpleInstrument"
            )
            with self.assertRaises(UserExpressionError):
                simpleQGraph.makeSimpleQGraph(root=root, pipeline=pipeline, userQuery="instrument = 'foo'")

    def testUserQueryBind(self):
        """Verify that bind values work for user query."""
        pipeline = simpleQGraph.makeSimplePipeline(
            nQuanta=5, instrument="lsst.pipe.base.tests.simpleQGraph.SimpleInstrument"
        )
        instr = simpleQGraph.SimpleInstrument.getName()
        # With a literal in the user query
        with temporaryDirectory() as root:
            simpleQGraph.makeSimpleQGraph(root=root, pipeline=pipeline, userQuery=f"instrument = '{instr}'")
        # With a bind for the user query
        with temporaryDirectory() as root:
            simpleQGraph.makeSimpleQGraph(
                root=root, pipeline=pipeline, userQuery="instrument = instr", bind={"instr": instr}
            )

    def test_datastore_records(self):
        """Test for generating datastore records."""
        with temporaryDirectory() as root:
            # need FileDatastore for this tests
            butler, qgraph1 = simpleQGraph.makeSimpleQGraph(
                root=root, inMemory=False, makeDatastoreRecords=True
            )

            # save and reload
            buffer = io.BytesIO()
            qgraph1.save(buffer)
            buffer.seek(0)
            qgraph2 = QuantumGraph.load(buffer, universe=butler.dimensions)
            del buffer

            for qgraph in (qgraph1, qgraph2):
                self.assertEqual(len(qgraph), 5)
                for i, qnode in enumerate(qgraph):
                    quantum = qnode.quantum
                    self.assertIsNotNone(quantum.datastore_records)
                    # only the first quantum has a pre-existing input
                    if i == 0:
                        datastore_name = "FileDatastore@<butlerRoot>"
                        self.assertEqual(set(quantum.datastore_records.keys()), {datastore_name})
                        records_data = quantum.datastore_records[datastore_name]
                        records = dict(records_data.records)
                        self.assertEqual(len(records), 1)
                        _, records = records.popitem()
                        records = records["file_datastore_records"]
                        self.assertEqual(
                            [record.path for record in records],
                            ["test/add_dataset0/add_dataset0_INSTR_det0_test.pickle"],
                        )
                    else:
                        self.assertEqual(quantum.datastore_records, {})


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining connection classes for PipelineTask.
"""

from __future__ import annotations

__all__ = [
    "AdjustQuantumHelper",
    "DeferredDatasetRef",
    "InputQuantizedConnection",
    "OutputQuantizedConnection",
    "PipelineTaskConnections",
    "ScalarError",
    "iterConnections",
    "ScalarError",
    "QuantizedConnection",
]

import dataclasses
import itertools
import string
import warnings
from collections import UserDict
from collections.abc import Collection, Generator, Iterable, Mapping, Sequence, Set
from dataclasses import dataclass
from types import MappingProxyType, SimpleNamespace
from typing import TYPE_CHECKING, Any

from lsst.daf.butler import DataCoordinate, DatasetRef, DatasetType, NamedKeyDict, NamedKeyMapping, Quantum

from ._status import NoWorkFound
from .connectionTypes import BaseConnection, BaseInput, Output, PrerequisiteInput

if TYPE_CHECKING:
    from .config import PipelineTaskConfig


class ScalarError(TypeError):
    """Exception raised when dataset type is configured as scalar
    but there are multiple data IDs in a Quantum for that dataset.
    """


class PipelineTaskConnectionDict(UserDict):
    """A special dict class used by `PipelineTaskConnectionMetaclass`.

    This dict is used in `PipelineTaskConnection` class creation, as the
    dictionary that is initially used as ``__dict__``. It exists to
    intercept connection fields declared in a `PipelineTaskConnection`, and
    what name is used to identify them. The names are then added to class
    level list according to the connection type of the class attribute. The
    names are also used as keys in a class level dictionary associated with
    the corresponding class attribute. This information is a duplicate of
    what exists in ``__dict__``, but provides a simple place to lookup and
    iterate on only these variables.

    Parameters
    ----------
    *args : `~typing.Any`
        Passed to `dict` constructor.
    **kwargs : `~typing.Any`
        Passed to `dict` constructor.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        # Initialize class level variables used to track any declared
        # class level variables that are instances of
        # connectionTypes.BaseConnection
        self.data["inputs"] = set()
        self.data["prerequisiteInputs"] = set()
        self.data["outputs"] = set()
        self.data["initInputs"] = set()
        self.data["initOutputs"] = set()
        self.data["allConnections"] = {}

    def __setitem__(self, name: str, value: Any) -> None:
        if isinstance(value, BaseConnection):
            if name in {
                "dimensions",
                "inputs",
                "prerequisiteInputs",
                "outputs",
                "initInputs",
                "initOutputs",
                "allConnections",
            }:
                # Guard against connections whose names are reserved.
                raise AttributeError(f"Connection name {name!r} is reserved for internal use.")
            if (previous := self.data.get(name)) is not None:
                # Guard against changing the type of an in inherited connection
                # by first removing it from the set it's current in.
                self.data[previous._connection_type_set].discard(name)
            object.__setattr__(value, "varName", name)
            self.data["allConnections"][name] = value
            self.data[value._connection_type_set].add(name)
        # defer to the default behavior
        super().__setitem__(name, value)


class PipelineTaskConnectionsMetaclass(type):
    """Metaclass used in the declaration of PipelineTaskConnections classes.

    Parameters
    ----------
    name : `str`
        Name of connection.
    bases : `~collections.abc.Collection`
        Base classes.
    dct : `~collections.abc.Mapping`
        Connections dict.
    **kwargs : `~typing.Any`
        Additional parameters.
    """

    # We can annotate these attributes as `collections.abc.Set` to discourage
    # undesirable modifications in type-checked code, since the internal code
    # modifying them is in `PipelineTaskConnectionDict` and that doesn't see
    # these annotations anyway.

    dimensions: Set[str]
    """Set of dimension names that define the unit of work for this task.

    Required and implied dependencies will automatically be expanded later and
    need not be provided.

    This is shadowed by an instance-level attribute on
    `PipelineTaskConnections` instances.
    """

    inputs: Set[str]
    """Set with the names of all `~connectionTypes.Input` connection
    attributes.

    This is updated automatically as class attributes are added.  Note that
    this attribute is shadowed by an instance-level attribute on
    `PipelineTaskConnections` instances.
    """

    prerequisiteInputs: Set[str]
    """Set with the names of all `~connectionTypes.PrerequisiteInput`
    connection attributes.

    See `inputs` for additional information.
    """

    outputs: Set[str]
    """Set with the names of all `~connectionTypes.Output` connection
    attributes.

    See `inputs` for additional information.
    """

    initInputs: Set[str]
    """Set with the names of all `~connectionTypes.InitInput` connection
    attributes.

    See `inputs` for additional information.
    """

    initOutputs: Set[str]
    """Set with the names of all `~connectionTypes.InitOutput` connection
    attributes.

    See `inputs` for additional information.
    """

    allConnections: Mapping[str, BaseConnection]
    """Mapping containing all connection attributes.

    See `inputs` for additional information.
    """

    def __prepare__(name, bases, **kwargs):  # noqa: N804
        # Create an instance of our special dict to catch and track all
        # variables that are instances of connectionTypes.BaseConnection
        # Copy any existing connections from a parent class
        dct = PipelineTaskConnectionDict()
        for base in bases:
            if isinstance(base, PipelineTaskConnectionsMetaclass):
                for name, value in base.allConnections.items():
                    dct[name] = value
        return dct

    def __new__(cls, name, bases, dct, **kwargs):
        dimensionsValueError = TypeError(
            "PipelineTaskConnections class must be created with a dimensions "
            "attribute which is an iterable of dimension names"
        )

        if name != "PipelineTaskConnections":
            # Verify that dimensions are passed as a keyword in class
            # declaration
            if "dimensions" not in kwargs:
                for base in bases:
                    if hasattr(base, "dimensions"):
                        kwargs["dimensions"] = base.dimensions
                        break
                if "dimensions" not in kwargs:
                    raise dimensionsValueError
            try:
                if isinstance(kwargs["dimensions"], str):
                    raise TypeError(
                        "Dimensions must be iterable of dimensions, got str,possibly omitted trailing comma"
                    )
                if not isinstance(kwargs["dimensions"], Iterable):
                    raise TypeError("Dimensions must be iterable of dimensions")
                dct["dimensions"] = set(kwargs["dimensions"])
            except TypeError as exc:
                raise dimensionsValueError from exc
            # Lookup any python string templates that may have been used in the
            # declaration of the name field of a class connection attribute
            allTemplates = set()
            stringFormatter = string.Formatter()
            # Loop over all connections
            for obj in dct["allConnections"].values():
                nameValue = obj.name
                # add all the parameters to the set of templates
                for param in stringFormatter.parse(nameValue):
                    if param[1] is not None:
                        allTemplates.add(param[1])

            # look up any template from base classes and merge them all
            # together
            mergeDict = {}
            mergeDeprecationsDict = {}
            for base in bases[::-1]:
                if hasattr(base, "defaultTemplates"):
                    mergeDict.update(base.defaultTemplates)
                if hasattr(base, "deprecatedTemplates"):
                    mergeDeprecationsDict.update(base.deprecatedTemplates)
            if "defaultTemplates" in kwargs:
                mergeDict.update(kwargs["defaultTemplates"])
            if "deprecatedTemplates" in kwargs:
                mergeDeprecationsDict.update(kwargs["deprecatedTemplates"])
            if len(mergeDict) > 0:
                kwargs["defaultTemplates"] = mergeDict
            if len(mergeDeprecationsDict) > 0:
                kwargs["deprecatedTemplates"] = mergeDeprecationsDict

            # Verify that if templated strings were used, defaults were
            # supplied as an argument in the declaration of the connection
            # class
            if len(allTemplates) > 0 and "defaultTemplates" not in kwargs:
                raise TypeError(
                    "PipelineTaskConnection class contains templated attribute names, but no "
                    "defaut templates were provided, add a dictionary attribute named "
                    "defaultTemplates which contains the mapping between template key and value"
                )
            if len(allTemplates) > 0:
                # Verify all templates have a default, and throw if they do not
                defaultTemplateKeys = set(kwargs["defaultTemplates"].keys())
                templateDifference = allTemplates.difference(defaultTemplateKeys)
                if templateDifference:
                    raise TypeError(f"Default template keys were not provided for {templateDifference}")
                # Verify that templates do not share names with variable names
                # used for a connection, this is needed because of how
                # templates are specified in an associated config class.
                nameTemplateIntersection = allTemplates.intersection(set(dct["allConnections"].keys()))
                if len(nameTemplateIntersection) > 0:
                    raise TypeError(
                        "Template parameters cannot share names with Class attributes"
                        f" (conflicts are {nameTemplateIntersection})."
                    )
            dct["defaultTemplates"] = kwargs.get("defaultTemplates", {})
            dct["deprecatedTemplates"] = kwargs.get("deprecatedTemplates", {})

        # Convert all the connection containers into frozensets so they cannot
        # be modified at the class scope
        for connectionName in ("inputs", "prerequisiteInputs", "outputs", "initInputs", "initOutputs"):
            dct[connectionName] = frozenset(dct[connectionName])
        # our custom dict type must be turned into an actual dict to be used in
        # type.__new__
        return super().__new__(cls, name, bases, dict(dct))

    def __init__(cls, name, bases, dct, **kwargs):
        # This overrides the default init to drop the kwargs argument. Python
        # metaclasses will have this argument set if any kwargs are passes at
        # class construction time, but should be consumed before calling
        # __init__ on the type metaclass. This is in accordance with python
        # documentation on metaclasses
        super().__init__(name, bases, dct)

    def __call__(cls, *, config: PipelineTaskConfig | None = None) -> PipelineTaskConnections:
        # MyPy appears not to really understand metaclass.__call__ at all, so
        # we need to tell it to ignore __new__ and __init__ calls here.
        instance: PipelineTaskConnections = cls.__new__(cls)  # type: ignore

        # Make mutable copies of all set-like class attributes so derived
        # __init__ implementations can modify them in place.
        instance.dimensions = set(cls.dimensions)
        instance.inputs = set(cls.inputs)
        instance.prerequisiteInputs = set(cls.prerequisiteInputs)
        instance.outputs = set(cls.outputs)
        instance.initInputs = set(cls.initInputs)
        instance.initOutputs = set(cls.initOutputs)

        # Set self.config.  It's a bit strange that we claim to accept None but
        # really just raise here, but it's not worth changing now.
        from .config import PipelineTaskConfig  # local import to avoid cycle

        if config is None or not isinstance(config, PipelineTaskConfig):
            raise ValueError(
                "PipelineTaskConnections must be instantiated with a PipelineTaskConfig instance"
            )
        instance.config = config

        # Extract the template names that were defined in the config instance
        # by looping over the keys of the defaultTemplates dict specified at
        # class declaration time.
        templateValues = {
            name: getattr(config.connections, name) for name in cls.defaultTemplates  # type: ignore
        }

        # We now assemble a mapping of all connection instances keyed by
        # internal name, applying the configuration and templates to make new
        # configurations from the class-attribute defaults.  This will be
        # private, but with a public read-only view.  This mapping is what the
        # descriptor interface of the class-level attributes will return when
        # they are accessed on an instance.  This is better than just assigning
        # regular instance attributes as it makes it so removed connections
        # cannot be accessed on instances, instead of having access to them
        # silent fall through to the not-removed class connection instance.
        instance._allConnections = {}
        instance.allConnections = MappingProxyType(instance._allConnections)
        for internal_name, connection in cls.allConnections.items():
            dataset_type_name = getattr(config.connections, internal_name).format(**templateValues)
            instance_connection = dataclasses.replace(
                connection,
                name=dataset_type_name,
                doc=(
                    connection.doc
                    if connection.deprecated is None
                    else f"{connection.doc}\n{connection.deprecated}"
                ),
                _deprecation_context=connection._deprecation_context,
            )
            instance._allConnections[internal_name] = instance_connection

        # Finally call __init__.  The base class implementation does nothing;
        # we could have left some of the above implementation there (where it
        # originated), but putting it here instead makes it hard for derived
        # class implementors to get things into a weird state by delegating to
        # super().__init__ in the wrong place, or by forgetting to do that
        # entirely.
        instance.__init__(config=config)  # type: ignore

        # Derived-class implementations may have changed the contents of the
        # various kinds-of-connection sets; update allConnections to have keys
        # that are a union of all those.  We get values for the new
        # allConnections from the attributes, since any dynamically added new
        # ones will not be present in the old allConnections.  Typically those
        # getattrs will invoke the descriptors and get things from the old
        # allConnections anyway.  After processing each set we replace it with
        # a frozenset.
        updated_all_connections = {}
        for attrName in ("initInputs", "prerequisiteInputs", "inputs", "initOutputs", "outputs"):
            updated_connection_names = getattr(instance, attrName)
            updated_all_connections.update(
                {name: getattr(instance, name) for name in updated_connection_names}
            )
            # Setting these to frozenset is at odds with the type annotation,
            # but MyPy can't tell because we're using setattr, and we want to
            # lie to it anyway to get runtime guards against post-__init__
            # mutation.
            setattr(instance, attrName, frozenset(updated_connection_names))
        # Update the existing dict in place, since we already have a view of
        # that.
        instance._allConnections.clear()
        instance._allConnections.update(updated_all_connections)

        for connection_name, obj in instance._allConnections.items():
            if obj.deprecated is not None:
                warnings.warn(
                    f"Connection {connection_name} with datasetType {obj.name} "
                    f"(from {obj._deprecation_context}): {obj.deprecated}",
                    FutureWarning,
                    stacklevel=1,  # Report from this location.
                )

        # Freeze the connection instance dimensions now.  This at odds with the
        # type annotation, which says [mutable] `set`, just like the connection
        # type attributes (e.g. `inputs`, `outputs`, etc.), though MyPy can't
        # tell with those since we're using setattr for them.
        instance.dimensions = frozenset(instance.dimensions)  # type: ignore

        return instance


class QuantizedConnection(SimpleNamespace):
    r"""A Namespace to map defined variable names of connections to the
    associated `lsst.daf.butler.DatasetRef` objects.

    This class maps the names used to define a connection on a
    `PipelineTaskConnections` class to the corresponding
    `~lsst.daf.butler.DatasetRef`\s provided by a `~lsst.daf.butler.Quantum`
    instance.  This will be a quantum of execution based on the graph created
    by examining all the connections defined on the
    `PipelineTaskConnections` class.

    Parameters
    ----------
    **kwargs : `~typing.Any`
        Not used.
    """

    def __init__(self, **kwargs):
        # Create a variable to track what attributes are added. This is used
        # later when iterating over this QuantizedConnection instance
        object.__setattr__(self, "_attributes", set())

    def __setattr__(self, name: str, value: DatasetRef | list[DatasetRef]) -> None:
        # Capture the attribute name as it is added to this object
        self._attributes.add(name)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        object.__delattr__(self, name)
        self._attributes.remove(name)

    def __len__(self) -> int:
        return len(self._attributes)

    def __iter__(
        self,
    ) -> Generator[tuple[str, DatasetRef | list[DatasetRef]], None, None]:
        """Make an iterator for this `QuantizedConnection`.

        Iterating over a `QuantizedConnection` will yield a tuple with the name
        of an attribute and the value associated with that name. This is
        similar to dict.items() but is on the namespace attributes rather than
        dict keys.
        """
        yield from ((name, getattr(self, name)) for name in self._attributes)

    def keys(self) -> Generator[str, None, None]:
        """Return an iterator over all the attributes added to a
        `QuantizedConnection` class.
        """
        yield from self._attributes


class InputQuantizedConnection(QuantizedConnection):
    """Input variant of a `QuantizedConnection`."""

    pass


class OutputQuantizedConnection(QuantizedConnection):
    """Output variant of a `QuantizedConnection`."""

    pass


@dataclass(frozen=True)
class DeferredDatasetRef:
    """A wrapper class for `~lsst.daf.butler.DatasetRef` that indicates that a
    `PipelineTask` should receive a `~lsst.daf.butler.DeferredDatasetHandle`
    instead of an in-memory dataset.

    Attributes
    ----------
    datasetRef : `lsst.daf.butler.DatasetRef`
        The `lsst.daf.butler.DatasetRef` that will be eventually used to
        resolve a dataset.
    """

    datasetRef: DatasetRef

    def __getattr__(self, name: str) -> Any:
        return getattr(self.datasetRef, name)


class PipelineTaskConnections(metaclass=PipelineTaskConnectionsMetaclass):
    """PipelineTaskConnections is a class used to declare desired IO when a
    PipelineTask is run by an activator.

    Parameters
    ----------
    config : `PipelineTaskConfig`
        A `PipelineTaskConfig` class instance whose class has been configured
        to use this `PipelineTaskConnections` class.

    See Also
    --------
    iterConnections : Iterator over selected connections.

    Notes
    -----
    ``PipelineTaskConnection`` classes are created by declaring class
    attributes of types defined in `lsst.pipe.base.connectionTypes` and are
    listed as follows:

    * ``InitInput`` - Defines connections in a quantum graph which are used as
      inputs to the ``__init__`` function of the `PipelineTask` corresponding
      to this class
    * ``InitOuput`` - Defines connections in a quantum graph which are to be
      persisted using a butler at the end of the ``__init__`` function of the
      `PipelineTask` corresponding to this class. The variable name used to
      define this connection should be the same as an attribute name on the
      `PipelineTask` instance. E.g. if an ``InitOutput`` is declared with
      the name ``outputSchema`` in a ``PipelineTaskConnections`` class, then
      a `PipelineTask` instance should have an attribute
      ``self.outputSchema`` defined. Its value is what will be saved by the
      activator framework.
    * ``PrerequisiteInput`` - An input connection type that defines a
      `lsst.daf.butler.DatasetType` that must be present at execution time,
      but that will not be used during the course of creating the quantum
      graph to be executed. These most often are things produced outside the
      processing pipeline, such as reference catalogs.
    * ``Input`` - Input `lsst.daf.butler.DatasetType` objects that will be used
      in the ``run`` method of a `PipelineTask`.  The name used to declare
      class attribute must match a function argument name in the ``run``
      method of a `PipelineTask`. E.g. If the ``PipelineTaskConnections``
      defines an ``Input`` with the name ``calexp``, then the corresponding
      signature should be ``PipelineTask.run(calexp, ...)``
    * ``Output`` - A `lsst.daf.butler.DatasetType` that will be produced by an
      execution of a `PipelineTask`. The name used to declare the connection
      must correspond to an attribute of a `Struct` that is returned by a
      `PipelineTask` ``run`` method.  E.g. if an output connection is
      defined with the name ``measCat``, then the corresponding
      ``PipelineTask.run`` method must return ``Struct(measCat=X,..)`` where
      X matches the ``storageClass`` type defined on the output connection.

    Attributes of these types can also be created, replaced, or deleted on the
    `PipelineTaskConnections` instance in the ``__init__`` method, if more than
    just the name depends on the configuration.  It is preferred to define them
    in the class when possible (even if configuration may cause the connection
    to be removed from the instance).

    The process of declaring a ``PipelineTaskConnection`` class involves
    parameters passed in the declaration statement.

    The first parameter is ``dimensions`` which is an iterable of strings which
    defines the unit of processing the run method of a corresponding
    `PipelineTask` will operate on. These dimensions must match dimensions that
    exist in the butler registry which will be used in executing the
    corresponding `PipelineTask`.  The dimensions may be also modified in
    subclass ``__init__`` methods if they need to depend on configuration.

    The second parameter is labeled ``defaultTemplates`` and is conditionally
    optional. The name attributes of connections can be specified as python
    format strings, with named format arguments. If any of the name parameters
    on connections defined in a `PipelineTaskConnections` class contain a
    template, then a default template value must be specified in the
    ``defaultTemplates`` argument. This is done by passing a dictionary with
    keys corresponding to a template identifier, and values corresponding to
    the value to use as a default when formatting the string. For example if
    ``ConnectionsClass.calexp.name = '{input}Coadd_calexp'`` then
    ``defaultTemplates`` = {'input': 'deep'}.

    Once a `PipelineTaskConnections` class is created, it is used in the
    creation of a `PipelineTaskConfig`. This is further documented in the
    documentation of `PipelineTaskConfig`. For the purposes of this
    documentation, the relevant information is that the config class allows
    configuration of connection names by users when running a pipeline.

    Instances of a `PipelineTaskConnections` class are used by the pipeline
    task execution framework to introspect what a corresponding `PipelineTask`
    will require, and what it will produce.

    Examples
    --------
    >>> from lsst.pipe.base import connectionTypes as cT
    >>> from lsst.pipe.base import PipelineTaskConnections
    >>> from lsst.pipe.base import PipelineTaskConfig
    >>> class ExampleConnections(PipelineTaskConnections,
    ...                          dimensions=("A", "B"),
    ...                          defaultTemplates={"foo": "Example"}):
    ...     inputConnection = cT.Input(doc="Example input",
    ...                                dimensions=("A", "B"),
    ...                                storageClass=Exposure,
    ...                                name="{foo}Dataset")
    ...     outputConnection = cT.Output(doc="Example output",
    ...                                  dimensions=("A", "B"),
    ...                                  storageClass=Exposure,
    ...                                  name="{foo}output")
    >>> class ExampleConfig(PipelineTaskConfig,
    ...                     pipelineConnections=ExampleConnections):
    ...    pass
    >>> config = ExampleConfig()
    >>> config.connections.foo = Modified
    >>> config.connections.outputConnection = "TotallyDifferent"
    >>> connections = ExampleConnections(config=config)
    >>> assert(connections.inputConnection.name == "ModifiedDataset")
    >>> assert(connections.outputConnection.name == "TotallyDifferent")
    """

    # We annotate these attributes as mutable sets because that's what they are
    # inside derived ``__init__`` implementations and that's what matters most
    # After that's done, the metaclass __call__ makes them into frozensets, but
    # relatively little code interacts with them then, and that code knows not
    # to try to modify them without having to be told that by mypy.

    dimensions: set[str]
    """Set of dimension names that define the unit of work for this task.

    Required and implied dependencies will automatically be expanded later and
    need not be provided.

    This may be replaced or modified in ``__init__`` to change the dimensions
    of the task.  After ``__init__`` it will be a `frozenset` and may not be
    replaced.
    """

    inputs: set[str]
    """Set with the names of all `connectionTypes.Input` connection attributes.

    This is updated automatically as class attributes are added, removed, or
    replaced in ``__init__``.  Removing entries from this set will cause those
    connections to be removed after ``__init__`` completes, but this is
    supported only for backwards compatibility; new code should instead just
    delete the collection attributed directly.  After ``__init__`` this will be
    a `frozenset` and may not be replaced.
    """

    prerequisiteInputs: set[str]
    """Set with the names of all `~connectionTypes.PrerequisiteInput`
    connection attributes.

    See `inputs` for additional information.
    """

    outputs: set[str]
    """Set with the names of all `~connectionTypes.Output` connection
    attributes.

    See `inputs` for additional information.
    """

    initInputs: set[str]
    """Set with the names of all `~connectionTypes.InitInput` connection
    attributes.

    See `inputs` for additional information.
    """

    initOutputs: set[str]
    """Set with the names of all `~connectionTypes.InitOutput` connection
    attributes.

    See `inputs` for additional information.
    """

    allConnections: Mapping[str, BaseConnection]
    """Mapping holding all connection attributes.

    This is a read-only view that is automatically updated when connection
    attributes are added, removed, or replaced in ``__init__``.  It is also
    updated after ``__init__`` completes to reflect changes in `inputs`,
    `prerequisiteInputs`, `outputs`, `initInputs`, and `initOutputs`.
    """

    _allConnections: dict[str, BaseConnection]

    def __init__(self, *, config: PipelineTaskConfig | None = None):
        pass

    def __setattr__(self, name: str, value: Any) -> None:
        if isinstance(value, BaseConnection):
            previous = self._allConnections.get(name)
            try:
                getattr(self, value._connection_type_set).add(name)
            except AttributeError:
                # Attempt to call add on a frozenset, which is what these sets
                # are after __init__ is done.
                raise TypeError("Connections objects are frozen after construction.") from None
            if previous is not None and value._connection_type_set != previous._connection_type_set:
                # Connection has changed type, e.g. Input to PrerequisiteInput;
                # update the sets accordingly. To be extra defensive about
                # multiple assignments we use the type of the previous instance
                # instead of assuming that's the same as the type of the self,
                # which is just the default.  Use discard instead of remove so
                # manually removing from these sets first is never an error.
                getattr(self, previous._connection_type_set).discard(name)
            self._allConnections[name] = value
            if hasattr(self.__class__, name):
                # Don't actually set the attribute if this was a connection
                # declared in the class; in that case we let the descriptor
                # return the value we just added to allConnections.
                return
        # Actually add the attribute.
        super().__setattr__(name, value)

    def __delattr__(self, name):
        """Descriptor delete method."""
        previous = self._allConnections.get(name)
        if previous is not None:
            # Delete this connection's name from the appropriate set, which we
            # have to get from the previous instance instead of assuming it's
            # the same set that was appropriate for the class-level default.
            # Use discard instead of remove so manually removing from these
            # sets first is never an error.
            try:
                getattr(self, previous._connection_type_set).discard(name)
            except AttributeError:
                # Attempt to call discard on a frozenset, which is what these
                # sets are after __init__ is done.
                raise TypeError("Connections objects are frozen after construction.") from None
            del self._allConnections[name]
            if hasattr(self.__class__, name):
                # Don't actually delete the attribute if this was a connection
                # declared in the class; in that case we let the descriptor
                # see that it's no longer present in allConnections.
                return
        # Actually delete the attribute.
        super().__delattr__(name)

    def buildDatasetRefs(
        self, quantum: Quantum
    ) -> tuple[InputQuantizedConnection, OutputQuantizedConnection]:
        """Build `QuantizedConnection` corresponding to input
        `~lsst.daf.butler.Quantum`.

        Parameters
        ----------
        quantum : `lsst.daf.butler.Quantum`
            Quantum object which defines the inputs and outputs for a given
            unit of processing.

        Returns
        -------
        retVal : `tuple` of (`InputQuantizedConnection`, \
                `OutputQuantizedConnection`)
            Namespaces mapping attribute names
            (identifiers of connections) to butler references defined in the
            input `~lsst.daf.butler.Quantum`.
        """
        inputDatasetRefs = InputQuantizedConnection()
        outputDatasetRefs = OutputQuantizedConnection()

        # populate inputDatasetRefs from quantum inputs
        for attributeName in itertools.chain(self.inputs, self.prerequisiteInputs):
            # get the attribute identified by name
            attribute = getattr(self, attributeName)
            # if the dataset is marked to load deferred, wrap it in a
            # DeferredDatasetRef
            quantumInputRefs: list[DatasetRef] | list[DeferredDatasetRef]
            if attribute.deferLoad:
                quantumInputRefs = [
                    DeferredDatasetRef(datasetRef=ref) for ref in quantum.inputs[attribute.name]
                ]
            else:
                quantumInputRefs = list(quantum.inputs[attribute.name])
            # Unpack arguments that are not marked multiples (list of
            # length one)
            if not attribute.multiple:
                if len(quantumInputRefs) > 1:
                    raise ScalarError(
                        "Received multiple datasets "
                        f"{', '.join(str(r.dataId) for r in quantumInputRefs)} "
                        f"for scalar connection {attributeName} "
                        f"({quantumInputRefs[0].datasetType.name}) "
                        f"of quantum for {quantum.taskName} with data ID {quantum.dataId}."
                    )
                if len(quantumInputRefs) == 0:
                    continue
                setattr(inputDatasetRefs, attributeName, quantumInputRefs[0])
            else:
                # Add to the QuantizedConnection identifier
                setattr(inputDatasetRefs, attributeName, quantumInputRefs)

        # populate outputDatasetRefs from quantum outputs
        for attributeName in self.outputs:
            # get the attribute identified by name
            attribute = getattr(self, attributeName)
            value = quantum.outputs[attribute.name]
            # Unpack arguments that are not marked multiples (list of
            # length one)
            if not attribute.multiple:
                setattr(outputDatasetRefs, attributeName, value[0])
            else:
                setattr(outputDatasetRefs, attributeName, value)

        return inputDatasetRefs, outputDatasetRefs

    def adjustQuantum(
        self,
        inputs: dict[str, tuple[BaseInput, Collection[DatasetRef]]],
        outputs: dict[str, tuple[Output, Collection[DatasetRef]]],
        label: str,
        data_id: DataCoordinate,
    ) -> tuple[
        Mapping[str, tuple[BaseInput, Collection[DatasetRef]]],
        Mapping[str, tuple[Output, Collection[DatasetRef]]],
    ]:
        """Override to make adjustments to `lsst.daf.butler.DatasetRef` objects
        in the `lsst.daf.butler.Quantum` during the graph generation stage
        of the activator.

        Parameters
        ----------
        inputs : `dict`
            Dictionary whose keys are an input (regular or prerequisite)
            connection name and whose values are a tuple of the connection
            instance and a collection of associated
            `~lsst.daf.butler.DatasetRef` objects.
            The exact type of the nested collections is unspecified; it can be
            assumed to be multi-pass iterable and support `len` and ``in``, but
            it should not be mutated in place.  In contrast, the outer
            dictionaries are guaranteed to be temporary copies that are true
            `dict` instances, and hence may be modified and even returned; this
            is especially useful for delegating to `super` (see notes below).
        outputs : `~collections.abc.Mapping`
            Mapping of output datasets, with the same structure as ``inputs``.
        label : `str`
            Label for this task in the pipeline (should be used in all
            diagnostic messages).
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for this quantum in the pipeline (should be used in all
            diagnostic messages).

        Returns
        -------
        adjusted_inputs : `~collections.abc.Mapping`
            Mapping of the same form as ``inputs`` with updated containers of
            input `~lsst.daf.butler.DatasetRef` objects.  Connections that are
            not changed should not be returned at all. Datasets may only be
            removed, not added.  Nested collections may be of any multi-pass
            iterable type, and the order of iteration will set the order of
            iteration within `PipelineTask.runQuantum`.
        adjusted_outputs : `~collections.abc.Mapping`
            Mapping of updated output datasets, with the same structure and
            interpretation as ``adjusted_inputs``.

        Raises
        ------
        ScalarError
            Raised if any `Input` or `PrerequisiteInput` connection has
            ``multiple`` set to `False`, but multiple datasets.
        NoWorkFound
            Raised to indicate that this quantum should not be run; not enough
            datasets were found for a regular `Input` connection, and the
            quantum should be pruned or skipped.
        FileNotFoundError
            Raised to cause QuantumGraph generation to fail (with the message
            included in this exception); not enough datasets were found for a
            `PrerequisiteInput` connection.

        Notes
        -----
        The base class implementation performs important checks.  It always
        returns an empty mapping (i.e. makes no adjustments).  It should
        always called be via `super` by custom implementations, ideally at the
        end of the custom implementation with already-adjusted mappings when
        any datasets are actually dropped, e.g.:

        .. code-block:: python

            def adjustQuantum(self, inputs, outputs, label, data_id):
                # Filter out some dataset refs for one connection.
                connection, old_refs = inputs["my_input"]
                new_refs = [ref for ref in old_refs if ...]
                adjusted_inputs = {"my_input", (connection, new_refs)}
                # Update the original inputs so we can pass them to super.
                inputs.update(adjusted_inputs)
                # Can ignore outputs from super because they are guaranteed
                # to be empty.
                super().adjustQuantum(inputs, outputs, label_data_id)
                # Return only the connections we modified.
                return adjusted_inputs, {}

        Removing outputs here is guaranteed to affect what is actually
        passed to `PipelineTask.runQuantum`, but its effect on the larger
        graph may be deferred to execution, depending on the context in
        which `adjustQuantum` is being run: if one quantum removes an output
        that is needed by a second quantum as input, the second quantum may not
        be adjusted (and hence pruned or skipped) until that output is actually
        found to be missing at execution time.

        Tasks that desire zip-iteration consistency between any combinations of
        connections that have the same data ID should generally implement
        `adjustQuantum` to achieve this, even if they could also run that
        logic during execution; this allows the system to see outputs that will
        not be produced because the corresponding input is missing as early as
        possible.
        """
        for name, (input_connection, refs) in inputs.items():
            dataset_type_name = input_connection.name
            if not input_connection.multiple and len(refs) > 1:
                raise ScalarError(
                    f"Found multiple datasets {', '.join(str(r.dataId) for r in refs)} "
                    f"for non-multiple input connection {label}.{name} ({dataset_type_name}) "
                    f"for quantum data ID {data_id}."
                )
            if len(refs) < input_connection.minimum:
                if isinstance(input_connection, PrerequisiteInput):
                    # This branch should only be possible during QG generation,
                    # or if someone deleted the dataset between making the QG
                    # and trying to run it.  Either one should be a hard error.
                    raise FileNotFoundError(
                        f"Not enough datasets ({len(refs)}) found for non-optional connection {label}.{name} "
                        f"({dataset_type_name}) with minimum={input_connection.minimum} for quantum data ID "
                        f"{data_id}."
                    )
                else:
                    raise NoWorkFound(label, name, input_connection)
        for name, (output_connection, refs) in outputs.items():
            dataset_type_name = output_connection.name
            if not output_connection.multiple and len(refs) > 1:
                raise ScalarError(
                    f"Found multiple datasets {', '.join(str(r.dataId) for r in refs)} "
                    f"for non-multiple output connection {label}.{name} ({dataset_type_name}) "
                    f"for quantum data ID {data_id}."
                )
        return {}, {}

    def getSpatialBoundsConnections(self) -> Iterable[str]:
        """Return the names of regular input and output connections whose data
        IDs should be used to compute the spatial bounds of this task's quanta.

        The spatial bound for a quantum is defined as the union of the regions
        of all data IDs of all connections returned here, along with the region
        of the quantum data ID (if the task has spatial dimensions).

        Returns
        -------
        connection_names : `collections.abc.Iterable` [ `str` ]
            Names of collections with spatial dimensions.  These are the
            task-internal connection names, not butler dataset type names.

        Notes
        -----
        The spatial bound is used to search for prerequisite inputs that have
        skypix dimensions. The default implementation returns an empty
        iterable, which is usually sufficient for tasks with spatial
        dimensions, but if a task's inputs or outputs are associated with
        spatial regions that extend beyond the quantum data ID's region, this
        method may need to be overridden to expand the set of prerequisite
        inputs found.

        Tasks that do not have spatial dimensions that have skypix prerequisite
        inputs should always override this method, as the default spatial
        bounds otherwise cover the full sky.
        """
        return ()

    def getTemporalBoundsConnections(self) -> Iterable[str]:
        """Return the names of regular input and output connections whose data
        IDs should be used to compute the temporal bounds of this task's
        quanta.

        The temporal bound for a quantum is defined as the union of the
        timespans of all data IDs of all connections returned here, along with
        the timespan of the quantum data ID (if the task has temporal
        dimensions).

        Returns
        -------
        connection_names : `collections.abc.Iterable` [ `str` ]
            Names of collections with temporal dimensions.  These are the
            task-internal connection names, not butler dataset type names.

        Notes
        -----
        The temporal bound is used to search for prerequisite inputs that are
        calibration datasets. The default implementation returns an empty
        iterable, which is usually sufficient for tasks with temporal
        dimensions, but if a task's inputs or outputs are associated with
        timespans that extend beyond the quantum data ID's timespan, this
        method may need to be overridden to expand the set of prerequisite
        inputs found.

        Tasks that do not have temporal dimensions that do not implement this
        method will use an infinite timespan for any calibration lookups.
        """
        return ()


def iterConnections(
    connections: PipelineTaskConnections, connectionType: str | Iterable[str]
) -> Generator[BaseConnection, None, None]:
    """Create an iterator over the selected connections type which yields
    all the defined connections of that type.

    Parameters
    ----------
    connections : `PipelineTaskConnections`
        An instance of a `PipelineTaskConnections` object that will be iterated
        over.
    connectionType : `str`
        The type of connections to iterate over, valid values are inputs,
        outputs, prerequisiteInputs, initInputs, initOutputs.

    Yields
    ------
    connection: `~.connectionTypes.BaseConnection`
        A connection defined on the input connections object of the type
        supplied.  The yielded value Will be an derived type of
        `~.connectionTypes.BaseConnection`.
    """
    if isinstance(connectionType, str):
        connectionType = (connectionType,)
    for name in itertools.chain.from_iterable(getattr(connections, ct) for ct in connectionType):
        yield getattr(connections, name)


@dataclass
class AdjustQuantumHelper:
    """Helper class for calling `PipelineTaskConnections.adjustQuantum`.

    This class holds `input` and `output` mappings in the form used by
    `Quantum` and execution harness code, i.e. with
    `~lsst.daf.butler.DatasetType` keys, translating them to and from the
    connection-oriented mappings used inside `PipelineTaskConnections`.
    """

    inputs: NamedKeyMapping[DatasetType, Sequence[DatasetRef]]
    """Mapping of regular input and prerequisite input datasets, grouped by
    `~lsst.daf.butler.DatasetType`.
    """

    outputs: NamedKeyMapping[DatasetType, Sequence[DatasetRef]]
    """Mapping of output datasets, grouped by `~lsst.daf.butler.DatasetType`.
    """

    inputs_adjusted: bool = False
    """Whether any inputs were removed in the last call to `adjust_in_place`.
    """

    outputs_adjusted: bool = False
    """Whether any outputs were removed in the last call to `adjust_in_place`.
    """

    def adjust_in_place(
        self,
        connections: PipelineTaskConnections,
        label: str,
        data_id: DataCoordinate,
    ) -> None:
        """Call `~PipelineTaskConnections.adjustQuantum` and update ``self``
        with its results.

        Parameters
        ----------
        connections : `PipelineTaskConnections`
            Instance on which to call `~PipelineTaskConnections.adjustQuantum`.
        label : `str`
            Label for this task in the pipeline (should be used in all
            diagnostic messages).
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID for this quantum in the pipeline (should be used in all
            diagnostic messages).
        """
        # Translate self's DatasetType-keyed, Quantum-oriented mappings into
        # connection-keyed, PipelineTask-oriented mappings.
        inputs_by_connection: dict[str, tuple[BaseInput, tuple[DatasetRef, ...]]] = {}
        outputs_by_connection: dict[str, tuple[Output, tuple[DatasetRef, ...]]] = {}
        for name in itertools.chain(connections.inputs, connections.prerequisiteInputs):
            connection = getattr(connections, name)
            dataset_type_name = connection.name
            inputs_by_connection[name] = (connection, tuple(self.inputs.get(dataset_type_name, ())))
        for name in itertools.chain(connections.outputs):
            connection = getattr(connections, name)
            dataset_type_name = connection.name
            outputs_by_connection[name] = (connection, tuple(self.outputs.get(dataset_type_name, ())))
        # Actually call adjustQuantum.
        # MyPy correctly complains that this call is not quite legal, but the
        # method docs explain exactly what's expected and it's the behavior we
        # want.  It'd be nice to avoid this if we ever have to change the
        # interface anyway, but not an immediate problem.
        adjusted_inputs_by_connection, adjusted_outputs_by_connection = connections.adjustQuantum(
            inputs_by_connection,  # type: ignore
            outputs_by_connection,  # type: ignore
            label,
            data_id,
        )
        # Translate adjustments to DatasetType-keyed, Quantum-oriented form,
        # installing new mappings in self if necessary.
        if adjusted_inputs_by_connection:
            adjusted_inputs = NamedKeyDict[DatasetType, tuple[DatasetRef, ...]](self.inputs)
            for name, (connection, updated_refs) in adjusted_inputs_by_connection.items():
                dataset_type_name = connection.name
                if not set(updated_refs).issubset(self.inputs[dataset_type_name]):
                    raise RuntimeError(
                        f"adjustQuantum implementation for task with label {label} returned {name} "
                        f"({dataset_type_name}) input datasets that are not a subset of those "
                        f"it was given for data ID {data_id}."
                    )
                adjusted_inputs[dataset_type_name] = tuple(updated_refs)
            self.inputs = adjusted_inputs.freeze()
            self.inputs_adjusted = True
        else:
            self.inputs_adjusted = False
        if adjusted_outputs_by_connection:
            adjusted_outputs = NamedKeyDict[DatasetType, tuple[DatasetRef, ...]](self.outputs)
            for name, (connection, updated_refs) in adjusted_outputs_by_connection.items():
                dataset_type_name = connection.name
                if not set(updated_refs).issubset(self.outputs[dataset_type_name]):
                    raise RuntimeError(
                        f"adjustQuantum implementation for task with label {label} returned {name} "
                        f"({dataset_type_name}) output datasets that are not a subset of those "
                        f"it was given for data ID {data_id}."
                    )
                adjusted_outputs[dataset_type_name] = tuple(updated_refs)
            self.outputs = adjusted_outputs.freeze()
            self.outputs_adjusted = True
        else:
            self.outputs_adjusted = False



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = (
    "NodeKey",
    "NodeType",
)

import enum
from typing import NamedTuple


class NodeType(enum.Enum):
    """Enumeration of the types of nodes in a PipelineGraph."""

    DATASET_TYPE = 0
    TASK_INIT = 1
    TASK = 2

    @property
    def bipartite(self) -> int:
        """The integer used as the "bipartite" key in networkx exports of a
        `PipelineGraph`.

        This key is used by the `networkx.algorithms.bipartite` module.
        """
        return int(self is not NodeType.DATASET_TYPE)

    def __lt__(self, other: NodeType) -> bool:
        # We define __lt__ only to be able to provide deterministic tiebreaking
        # on top of topological ordering of `PipelineGraph`` and views thereof.
        return self.value < other.value


class NodeKey(NamedTuple):
    """A special key type for nodes in networkx graphs.

    Notes
    -----
    Using a tuple for the key allows tasks labels and dataset type names with
    the same string value to coexist in the graph.  These only rarely appear in
    `PipelineGraph` public interfaces; when the node type is implicit, bare
    `str` task labels or dataset type names are used instead.

    NodeKey objects stringify to just their name, which is used both as a way
    to convert to the `str` objects used in the main public interface and as an
    easy way to usefully stringify containers returned directly by networkx
    algorithms (especially in error messages).  Note that this requires `repr`,
    not just `str`, because Python builtin containers always use `repr` on
    their items, even in their implementations for `str`.
    """

    node_type: NodeType
    """Node type enum for this key."""

    name: str
    """Task label or dataset type name.

    This is always the parent dataset type name for component dataset types.
    """

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.name

    @property
    def node_id(self) -> str:
        return f"{self.name}:{self.node_type.value}"
